import warnings
from pydm.display import Display, ScreenTarget, load_file, load_py_file, load_ui_file

warnings.warn(
    "The display_module was renamed to display." "Please modify your code to reflect this change.",
    DeprecationWarning,
    stacklevel=2,
)

__all__ = ["Display", "ScreenTarget", "load_file", "load_py_file", "load_ui_file"]



from ...widgets.eventplot import PyDMEventPlot


def test_add_channel(qtbot):
    """A quick check to ensure adding a channel to an event plot works as expected"""
    event_plot = PyDMEventPlot()
    qtbot.addWidget(event_plot)

    curve = "TEST:EVENT:PLOT"
    event_plot.addChannel(curve)

    assert event_plot.curveAtIndex(0).channel.address == "TEST:EVENT:PLOT"



import os
import tempfile
import pytest

from ...utilities.macro import substitute_in_file, parse_macro_string


@pytest.mark.parametrize(
    "text, macros, expected",
    [
        (
            "This is a ${what} to ensure that macros work. ${A} ${B} ${C}",
            {"what": "test", "A": "X", "B": "Y", "C": "Z"},
            "This is a test to ensure that macros work. X Y Z",
        ),
        (
            "This is a ${what} to ensure that macros work. ${A} ${B} ${C}",
            {"what": "test", "A": "X", "B": "Y", "C": "Z", "D": "W"},
            "This is a test to ensure that macros work. X Y Z",
        ),
        (
            "This is a ${what} to ensure that macros work. ${A} ${B} ${C}",
            {"what": "test", "A": "X", "B": "Y"},
            "This is a test to ensure that macros work. X Y ${C}",
        ),
    ],
)
def test_substitute_in_file(text, macros, expected):
    fd, name = tempfile.mkstemp(text=True)
    # use a context manager to open the file at that path and close it again
    with open(name, "w") as f:
        f.write(text)
    # close the file descriptor
    os.close(fd)

    nf = substitute_in_file(name, macros)
    nt = nf.read()
    assert nt == expected


@pytest.mark.parametrize(
    "macro_string, expected_dict",
    [
        ('{"A": "1", "B": "2"}', {"A": "1", "B": "2"}),
        ("A=1,B=2", {"A": "1", "B": "2"}),
        ("A=$(other_macro),B=2,C=3", {"A": "$(other_macro)", "B": "2", "C": "3"}),
        ("A=$(other_macro=3)", {"A": "$(other_macro=3)"}),
        ("TITLE='1,2', B=2, C=3", {"TITLE": "1,2", "B": "2", "C": "3"}),
        ("TITLE=1\,2,B=2,C=3", {"TITLE": "1,2", "B": "2", "C": "3"}),
        ('TITLE="e=mc^2",B=2,C=3', {"TITLE": "e=mc^2", "B": "2", "C": "3"}),
        ("", {}),
        (None, {}),
    ],
)
def test_macro_parser(macro_string, expected_dict):
    """
    Test the parser, using a couple of normal cases, and a bunch of perverse
    macro strings that only some insane macro genius (or huge macro idiot)
    would ever attempt.
    """
    assert parse_macro_string(macro_string) == expected_dict



import pytest
from ...utilities import iconfont
from qtpy import QtGui, QtCore


def test_icon_font_constructor(qtbot):
    icon_f = iconfont.IconFont()
    icon_f2 = iconfont.IconFont()
    assert icon_f is icon_f2


def test_icon_font_load_font(qtbot):
    icon_f = iconfont.IconFont()
    with pytest.raises(OSError):
        icon_f.char_map = None
        icon_f.load_font("foo", icon_f.charmap_file)
    with pytest.raises(OSError):
        icon_f.char_map = None
        icon_f.load_font(icon_f.charmap_file, "foo")
    icon_f.load_font(icon_f.font_file, icon_f.charmap_file)
    assert icon_f.char_map is not None


def test_icon_font_get_char_for_name(qtbot):
    icon_f = iconfont.IconFont()
    c = icon_f.get_char_for_name("cogs")
    assert c == "\uf085"

    with pytest.raises(ValueError):
        icon_f.get_char_for_name("foo")


def test_icon_font_font(qtbot):
    icon_f = iconfont.IconFont()
    f = icon_f.font(12)
    assert f.family() == icon_f.font_name
    assert f.pixelSize() == 12


def test_icon_font_icon(qtbot):
    icon_f = iconfont.IconFont()
    icon_f.icon("cogs", color=None)
    icon_f.icon("cogs", color=QtGui.QColor(255, 0, 0))
    with pytest.raises(ValueError):
        icon_f.icon("foo", color=None)


def test_char_icon_engine(qtbot):
    engine = iconfont.CharIconEngine(iconfont.IconFont(), "cogs", color=None)
    engine.pixmap(QtCore.QSize(32, 32), mode=QtGui.QIcon.Normal, state=QtGui.QIcon.On)
    engine.pixmap(QtCore.QSize(32, 32), mode=QtGui.QIcon.Disabled, state=QtGui.QIcon.On)



import pytest

from ...utilities import colors


def test_read_file():
    assert colors.svg_color_to_hex_map is not None
    assert colors.hex_to_svg_color_map is not None


def test_svg_color_from_hex():
    svg = colors.svg_color_from_hex("#000000", hex_on_fail=False)
    assert svg == "black"
    svg = colors.svg_color_from_hex("#000000", hex_on_fail=True)
    assert svg == "black"

    with pytest.raises(KeyError):
        colors.svg_color_from_hex("#XXXXXXX", hex_on_fail=False)
    svg = colors.svg_color_from_hex("#XXXXXXX", hex_on_fail=True)
    assert svg == "#XXXXXXX"


def test_hex_from_svg_color():
    hex = colors.hex_from_svg_color("black")
    assert hex == "#000000"

    with pytest.raises(KeyError):
        colors.hex_from_svg_color("invalid_color")



import pytest

from qtpy.QtGui import QColor
from qtpy.QtCore import Qt
from ...widgets.byte import PyDMMultiStateIndicator


@pytest.mark.parametrize(
    "value, expectedColor",
    [
        (0, QColor(Qt.black)),
        (4, QColor(Qt.red)),
        (7, QColor(Qt.darkGreen)),
        (15, QColor(Qt.yellow)),
    ],
)
def test_state_change(qtbot, signals, value, expectedColor):
    """
    Test the widget's handling of the value changed event.

    Expectations:
    1. Widget state is update to reflect incoming signal value
    2. Widgets color after state-update is the correct color set for the new state

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget testing
    signals : fixture
        The signals fixture, which provides access signals to be bound to the appropriate slots
    expected : int
        Expected resulting color after state-update
    """
    pydm_multistate = PyDMMultiStateIndicator()
    qtbot.addWidget(pydm_multistate)
    pydm_multistate.show()

    pydm_multistate.state4Color = QColor(Qt.red)
    pydm_multistate.state7Color = QColor(Qt.darkGreen)
    pydm_multistate.state15Color = QColor(Qt.yellow)

    pydm_multistate._connected = True

    signals.new_value_signal[type(value)].connect(pydm_multistate.channelValueChanged)
    signals.new_value_signal[type(value)].emit(value)

    assert pydm_multistate._curr_state == value
    assert pydm_multistate._curr_color == expectedColor



import logging
import os

from ..utilities.iconfont import IconFont
from .archiver_time_plot import PyDMArchiverTimePlot
from .byte import PyDMByteIndicator
from .byte import PyDMMultiStateIndicator
from .checkbox import PyDMCheckbox
from .datetime import PyDMDateTimeEdit, PyDMDateTimeLabel
from .drawing import (
    PyDMDrawingArc,
    PyDMDrawingChord,
    PyDMDrawingCircle,
    PyDMDrawingEllipse,
    PyDMDrawingImage,
    PyDMDrawingIrregularPolygon,
    PyDMDrawingLine,
    PyDMDrawingPie,
    PyDMDrawingPolygon,
    PyDMDrawingPolyline,
    PyDMDrawingRectangle,
    PyDMDrawingTriangle,
)
from .embedded_display import PyDMEmbeddedDisplay
from .enum_button import PyDMEnumButton
from .enum_combo_box import PyDMEnumComboBox
from .frame import PyDMFrame
from .image import PyDMImageView
from .label import PyDMLabel
from .line_edit import PyDMLineEdit
from .logdisplay import PyDMLogDisplay
from .pushbutton import PyDMPushButton
from .qtplugin_base import WidgetCategory, get_widgets_from_entrypoints, qtplugin_factory
from .qtplugin_extensions import (
    ArchiveTimeCurveEditorExtension,
    BasicSettingsExtension,
    RulesExtension,
    ScatterCurveEditorExtension,
    EventCurveEditorExtension,
    SymbolExtension,
    TimeCurveEditorExtension,
    WaveformCurveEditorExtension,
)
from .related_display_button import PyDMRelatedDisplayButton
from .scale import PyDMScaleIndicator
from .scatterplot import PyDMScatterPlot
from .shell_command import PyDMShellCommand
from .slider import PyDMSlider
from .spinbox import PyDMSpinbox
from .symbol import PyDMSymbol
from .waveformtable import PyDMWaveformTable
from .analog_indicator import PyDMAnalogIndicator
from .timeplot import PyDMTimePlot
from .waveformplot import PyDMWaveformPlot
from .eventplot import PyDMEventPlot
from .tab_bar_qtplugin import TabWidgetPlugin
from .template_repeater import PyDMTemplateRepeater
from .terminator import PyDMTerminator
from .nt_table import PyDMNTTable

logger = logging.getLogger(__name__)

ifont = IconFont()

BASE_EXTENSIONS = [BasicSettingsExtension, RulesExtension]


# Label plugin
PyDMLabelPlugin = qtplugin_factory(
    PyDMLabel, group=WidgetCategory.DISPLAY, extensions=BASE_EXTENSIONS, icon=ifont.icon("tag")
)

# Time Plot plugin
PyDMTimePlotPlugin = qtplugin_factory(
    PyDMTimePlot,
    group=WidgetCategory.PLOT,
    extensions=[TimeCurveEditorExtension, RulesExtension],
    icon=ifont.icon("chart-line"),
)

# In order to keep the archiver functionality invisible to users who do not have access to an instance of the
# archiver appliance, only load this if the user has the associated environment variable set
if "PYDM_ARCHIVER_URL" in os.environ:
    # Time Plot with archiver appliance support plugin
    PyDMArchiverTimePlotPlugin = qtplugin_factory(
        PyDMArchiverTimePlot,
        group=WidgetCategory.PLOT,
        extensions=[ArchiveTimeCurveEditorExtension, RulesExtension],
        icon=ifont.icon("chart-line"),
    )

# Waveform Plot plugin
PyDMWaveformPlotPlugin = qtplugin_factory(
    PyDMWaveformPlot,
    group=WidgetCategory.PLOT,
    extensions=[WaveformCurveEditorExtension, RulesExtension],
    icon=ifont.icon("wave-square"),
)

# Scatter Plot plugin
PyDMScatterPlotPlugin = qtplugin_factory(
    PyDMScatterPlot,
    group=WidgetCategory.PLOT,
    extensions=[ScatterCurveEditorExtension, RulesExtension],
    icon=ifont.icon("project-diagram"),
)

# Event Plot plugin
PyDMEventPlotPlugin = qtplugin_factory(
    PyDMEventPlot,
    group=WidgetCategory.PLOT,
    extensions=[EventCurveEditorExtension, RulesExtension],
    icon=ifont.icon("project-diagram"),
)

# Byte plugin
PyDMByteIndicatorPlugin = qtplugin_factory(
    PyDMByteIndicator, group=WidgetCategory.DISPLAY, extensions=BASE_EXTENSIONS, icon=ifont.icon("ellipsis-v")
)

# Multi-state plugin
PyDMMultiStateLEDIndicatorPlugin = qtplugin_factory(
    PyDMMultiStateIndicator, group=WidgetCategory.DISPLAY, extensions=BASE_EXTENSIONS, icon=ifont.icon("ellipsis-v")
)

# Checkbox plugin
PyDMCheckboxPlugin = qtplugin_factory(
    PyDMCheckbox, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("check-square")
)

# Date/Time plugins
PyDMDateTimeEditPlugin = qtplugin_factory(
    PyDMDateTimeEdit, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("calendar-minus")
)

PyDMDateTimeLabelPlugin = qtplugin_factory(
    PyDMDateTimeLabel, group=WidgetCategory.DISPLAY, extensions=BASE_EXTENSIONS, icon=ifont.icon("calendar-alt")
)
# Drawing plugins
PyDMDrawingArcPlugin = qtplugin_factory(
    PyDMDrawingArc, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("circle-notch")
)
PyDMDrawingChordPlugin = qtplugin_factory(
    PyDMDrawingChord, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("moon")
)
PyDMDrawingCirclePlugin = qtplugin_factory(
    PyDMDrawingCircle, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("circle")
)
PyDMDrawingEllipsePlugin = qtplugin_factory(
    PyDMDrawingEllipse, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("ellipsis-h")
)
PyDMDrawingImagePlugin = qtplugin_factory(
    PyDMDrawingImage, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("image")
)
PyDMDrawingLinePlugin = qtplugin_factory(
    PyDMDrawingLine, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("minus")
)
PyDMDrawingPiePlugin = qtplugin_factory(
    PyDMDrawingPie, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("pizza-slice")
)

PyDMDrawingRectanglePlugin = qtplugin_factory(
    PyDMDrawingRectangle, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("border-style")
)
PyDMDrawingTrianglePlugin = qtplugin_factory(
    PyDMDrawingTriangle, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("caret-up")
)

PyDMDrawingPolygonPlugin = qtplugin_factory(
    PyDMDrawingPolygon, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("draw-polygon")
)

PyDMDrawingPolylinePlugin = qtplugin_factory(
    PyDMDrawingPolyline, group=WidgetCategory.DRAWING, extensions=BASE_EXTENSIONS, icon=ifont.icon("share-alt")
)

PyDMDrawingIrregularPolygonPlugin = qtplugin_factory(
    PyDMDrawingIrregularPolygon,
    group=WidgetCategory.DRAWING,
    extensions=BASE_EXTENSIONS,
    icon=ifont.icon("draw-polygon"),
)

# Embedded Display plugin
PyDMEmbeddedDisplayPlugin = qtplugin_factory(
    PyDMEmbeddedDisplay, group=WidgetCategory.CONTAINER, extensions=BASE_EXTENSIONS, icon=ifont.icon("layer-group")
)

# Enum Button plugin
PyDMEnumButtonPlugin = qtplugin_factory(
    PyDMEnumButton, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("bars")
)

# Enum Combobox plugin
PyDMEnumComboBoxPlugin = qtplugin_factory(
    PyDMEnumComboBox, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("list-ol")
)

# Frame plugin
PyDMFramePlugin = qtplugin_factory(
    PyDMFrame, group=WidgetCategory.CONTAINER, is_container=True, extensions=BASE_EXTENSIONS, icon=ifont.icon("expand")
)

# Image plugin
PyDMImageViewPlugin = qtplugin_factory(
    PyDMImageView, group=WidgetCategory.DISPLAY, extensions=BASE_EXTENSIONS, icon=ifont.icon("camera")
)

# Line Edit plugin
PyDMLineEditPlugin = qtplugin_factory(
    PyDMLineEdit, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("edit")
)

# Log Viewer
PyDMLogDisplayPlugin = qtplugin_factory(
    PyDMLogDisplay, group=WidgetCategory.DISPLAY, extensions=BASE_EXTENSIONS, icon=ifont.icon("clipboard")
)

# Push Button plugin
PyDMPushButtonPlugin = qtplugin_factory(
    PyDMPushButton, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("mouse")
)

# Related Display Button plugin
PyDMRelatedDisplayButtonPlugin = qtplugin_factory(
    PyDMRelatedDisplayButton,
    group=WidgetCategory.DISPLAY,
    extensions=BASE_EXTENSIONS,
    icon=ifont.icon("window-maximize"),
)

# Shell Command plugin
PyDMShellCommandPlugin = qtplugin_factory(
    PyDMShellCommand, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("terminal")
)

# Slider plugin
PyDMSliderPlugin = qtplugin_factory(
    PyDMSlider, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("sliders-h")
)

# Spinbox plugin
PyDMSpinboxplugin = qtplugin_factory(
    PyDMSpinbox, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("sort-numeric-up")
)

# Scale Indicator plugin
PyDMScaleIndicatorPlugin = qtplugin_factory(
    PyDMScaleIndicator, group=WidgetCategory.DISPLAY, extensions=BASE_EXTENSIONS, icon=ifont.icon("level-up-alt")
)

# Analog Indicator plugin
PyDMAnalogIndicatorPlugin = qtplugin_factory(
    PyDMAnalogIndicator, group=WidgetCategory.DISPLAY, extensions=BASE_EXTENSIONS, icon=ifont.icon("level-up-alt")
)

# Symbol plugin
PyDMSymbolPlugin = qtplugin_factory(
    PyDMSymbol, group=WidgetCategory.DISPLAY, extensions=[SymbolExtension, RulesExtension], icon=ifont.icon("icons")
)

# Waveform Table plugin
PyDMWaveformTablePlugin = qtplugin_factory(
    PyDMWaveformTable, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("table")
)
# NTTable plugin
PyDMNTTable = qtplugin_factory(
    PyDMNTTable, group=WidgetCategory.INPUT, extensions=BASE_EXTENSIONS, icon=ifont.icon("table")
)

# Tab Widget plugin
PyDMTabWidgetPlugin = TabWidgetPlugin(extensions=BASE_EXTENSIONS)

# Template Repeater plugin
PyDMTemplateRepeaterPlugin = qtplugin_factory(
    PyDMTemplateRepeater, group=WidgetCategory.CONTAINER, extensions=BASE_EXTENSIONS, icon=ifont.icon("align-justify")
)

# Terminator Widget plugin
PyDMTerminatorPlugin = qtplugin_factory(PyDMTerminator, group=WidgetCategory.MISC, extensions=BASE_EXTENSIONS)

# **********************************************
# NOTE: Add in new PyDM widgets above this line.
# **********************************************

# Add in designer widget plugins from other classes via entrypoints:
globals().update(**get_widgets_from_entrypoints())



# Unit Tests for the Channel widget class
import pytest

from ...widgets.label import PyDMLabel
from ...widgets.line_edit import PyDMLineEdit
from ...widgets.channel import PyDMChannel
from pydm.data_plugins import plugin_for_address


class A:
    pass


def test_construct(qtbot):
    """
    Test the construct of the widget.

    Expectations:
    1. The widget is created with all the default values for its properties
    2. When created for a specific widget type, the Channel widget contains the default values, i.e. the appropriate
        slots, for that specific widget type
    3. Channel widgets can be compared for equality and inequality using the '==' and '!=' operators.

    Parameters
    ----------
    qtbot : fixture
        Window for widget testing
    """
    pydm_channel = PyDMChannel()

    assert (
        pydm_channel.address is None
        and pydm_channel.connection_slot is None
        and pydm_channel.value_slot is None
        and pydm_channel.severity_slot is None
        and pydm_channel.enum_strings_slot is None
        and pydm_channel.unit_slot is None
        and pydm_channel.prec_slot is None
        and pydm_channel.upper_ctrl_limit_slot is None
        and pydm_channel.lower_ctrl_limit_slot is None
        and pydm_channel.upper_alarm_limit_slot is None
        and pydm_channel.lower_alarm_limit_slot is None
        and pydm_channel.upper_warning_limit_slot is None
        and pydm_channel.lower_warning_limit_slot is None
        and pydm_channel.write_access_slot is None
        and pydm_channel.value_signal is None
        and pydm_channel.timestamp_slot is None
    )

    pydm_label = PyDMLabel(init_channel="tst://this")
    qtbot.addWidget(pydm_label)

    pydm_label_channels = pydm_label.channels()[0]
    default_pydm_label_channels = PyDMChannel(
        address=pydm_label.channel,
        connection_slot=pydm_label.connectionStateChanged,
        value_slot=pydm_label.channelValueChanged,
        severity_slot=pydm_label.alarmSeverityChanged,
        enum_strings_slot=pydm_label.enumStringsChanged,
        unit_slot=pydm_label.unitChanged,
        prec_slot=pydm_label.precisionChanged,
        upper_ctrl_limit_slot=pydm_label.upperCtrlLimitChanged,
        lower_ctrl_limit_slot=pydm_label.lowerCtrlLimitChanged,
        upper_alarm_limit_slot=pydm_label.upper_alarm_limit_changed,
        lower_alarm_limit_slot=pydm_label.lower_alarm_limit_changed,
        upper_warning_limit_slot=pydm_label.upper_warning_limit_changed,
        lower_warning_limit_slot=pydm_label.lower_warning_limit_changed,
        value_signal=None,
        write_access_slot=None,
        timestamp_slot=pydm_label.timestamp_changed,
    )
    assert pydm_label_channels == default_pydm_label_channels

    pydm_lineedit = PyDMLineEdit(init_channel="tst://this2")
    qtbot.addWidget(pydm_lineedit)

    # Test equal and not equal comparisons
    pydm_lineedit_channels = pydm_lineedit.channels()[0]
    assert pydm_lineedit_channels != default_pydm_label_channels

    not_same_type = A()
    equal_result = not_same_type == default_pydm_label_channels
    not_equal_result = not_same_type != default_pydm_label_channels
    assert equal_result is False and not_equal_result is True


def test_pydm_connection(test_plugin):
    # Plugin, Channel and Registry
    chan = PyDMChannel("tst://Tst:this3")
    plugin = plugin_for_address(chan.address)
    plugin_no = len(plugin.connections)

    print("Connections Before:")
    for k, v in plugin.connections.items():
        print("\t", k, " - ", v)

    # Make a connection
    chan.connect()

    print("Connections After:")
    for k, v in plugin.connections.items():
        print("\t", k, " - ", v)

    assert len(plugin.connections) == plugin_no + 1
    # Remove connections
    chan.disconnect()
    assert len(plugin.connections) == plugin_no


@pytest.mark.parametrize(
    "ch, ch_expected",
    [
        ("ca://MTEST:Float", "ca://MTEST:Float"),
        (" foo://bar", "foo://bar"),
        (" foo://bar ", "foo://bar"),
        ("foo://bar ", "foo://bar"),
        ("\nfoo://bar", "foo://bar"),
        ("\tfoo://bar", "foo://bar"),
        ("", ""),
        (None, None),
    ],
)
def test_channel_address(ch, ch_expected):
    channel = PyDMChannel()
    channel.address = ch
    assert channel.address == ch_expected



import argparse
import cProfile
import logging
import os
import pstats
import sys
import faulthandler
from qtpy import QtCore, QtGui


def main():
    logger = logging.getLogger("")
    handler = logging.StreamHandler()
    formatter = logging.Formatter("[%(asctime)s] [%(levelname)-8s] - %(message)s")
    handler.setFormatter(formatter)
    logger.addHandler(handler)
    logger.setLevel("INFO")
    handler.setLevel("INFO")

    from pydm.utilities import setup_renderer

    setup_renderer()

    try:
        """
        We must import QtWebEngineWidgets before creating a QApplication
        otherwise we get the following error if someone adds a WebView at Designer:
        ImportError: QtWebEngineWidgets must be imported before a QCoreApplication instance is created
        """
        from qtpy import QtWebEngineWidgets  # noqa: F401
    except ImportError:
        logger.debug("QtWebEngine is not supported.")

    import pydm
    from pydm.utilities.macro import parse_macro_string

    parser = argparse.ArgumentParser(description="Python Display Manager")
    parser.add_argument(
        "displayfile",
        help="A PyDM file to display." + "    Can be either a Qt .ui file, or a Python file.",
        nargs="?",
        default=None,
    )
    parser.add_argument(
        "--homefile", help="Path to a PyDM file to return to when the home button is clicked in the navigation bar"
    )
    parser.add_argument(
        "--perfmon",
        action="store_true",
        help="Enable performance monitoring," + " and print CPU usage to the terminal.",
    )
    parser.add_argument("--profile", action="store_true", help="Enable cProfile function profiling, printing on exit.")
    parser.add_argument("--faulthandler", action="store_true", help="Enable faulthandler to trace segmentation faults.")
    parser.add_argument("--hide-nav-bar", action="store_true", help="Start PyDM with the navigation bar hidden.")
    parser.add_argument("--hide-menu-bar", action="store_true", help="Start PyDM with the menu bar hidden.")
    parser.add_argument("--hide-status-bar", action="store_true", help="Start PyDM with the status bar hidden.")
    parser.add_argument("--fullscreen", action="store_true", help="Start PyDM in full screen mode.")
    parser.add_argument("--read-only", action="store_true", help="Start PyDM in a Read-Only mode.")
    parser.add_argument(
        "--log_level",
        help="Configure level of log display",
        choices=["DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"],
        default="INFO",
    )
    parser.add_argument(
        "--version",
        action="version",
        version="PyDM {version}".format(version=pydm.__version__),
        help="Show PyDM's version number and exit.",
    )
    parser.add_argument(
        "-m",
        "--macro",
        help="Specify macro replacements to use, in JSON object format."
        + "    Reminder: JSON requires double quotes for strings, "
        + "so you should wrap this whole argument in single quotes."
        + '  Example: -m \'{"sector": "LI25", "facility": "LCLS"}\''
        + "--or-- specify macro replacements as KEY=value pairs "
        + " using a comma as delimiter  If you want to uses spaces "
        + " after the delimiters or around the = signs, "
        + " wrap the entire set with quotes "
        + '  Example: -m "sector = LI25, facility=LCLS"',
    )
    parser.add_argument(
        "--stylesheet",
        help="Specify the full path to a CSS stylesheet file, which"
        + " can be used to customize the appearance of PyDM and"
        + " Qt widgets.",
        default=None,
    )
    parser.add_argument(
        "display_args",
        help="Arguments to be passed to the PyDM client application" + " (which is a QApplication subclass).",
        nargs=argparse.REMAINDER,
        default=None,
    )

    pydm_args = parser.parse_args()
    if pydm_args.profile:
        profile = cProfile.Profile()
        profile.enable()

    if pydm_args.faulthandler:
        faulthandler.enable()

    macros = None
    if pydm_args.macro is not None:
        macros = parse_macro_string(pydm_args.macro)

    if pydm_args.log_level:
        logger.setLevel(pydm_args.log_level)
        handler.setLevel(pydm_args.log_level)

    app = pydm.PyDMApplication(
        ui_file=pydm_args.displayfile,
        command_line_args=pydm_args.display_args,
        perfmon=pydm_args.perfmon,
        hide_nav_bar=pydm_args.hide_nav_bar,
        hide_menu_bar=pydm_args.hide_menu_bar,
        hide_status_bar=pydm_args.hide_status_bar,
        fullscreen=pydm_args.fullscreen,
        read_only=pydm_args.read_only,
        macros=macros,
        stylesheet_path=pydm_args.stylesheet,
        home_file=pydm_args.homefile,
    )

    base_path = os.path.dirname(os.path.realpath(__file__))
    icon_path_mask = os.path.join(base_path, "icons", "pydm_{}.png")

    app_icon = QtGui.QIcon()
    app_icon.addFile(icon_path_mask.format(16), QtCore.QSize(16, 16))
    app_icon.addFile(icon_path_mask.format(24), QtCore.QSize(24, 24))
    app_icon.addFile(icon_path_mask.format(32), QtCore.QSize(32, 32))
    app_icon.addFile(icon_path_mask.format(64), QtCore.QSize(64, 64))
    app_icon.addFile(icon_path_mask.format(128), QtCore.QSize(128, 128))
    app_icon.addFile(icon_path_mask.format(256), QtCore.QSize(256, 256))

    app.setWindowIcon(app_icon)
    app.setApplicationName("PyDM")

    pydm.utilities.shortcuts.install_connection_inspector(parent=app.main_window)

    exit_code = app.exec_()

    if pydm_args.profile:
        profile.disable()
        stats = pstats.Stats(
            profile,
            stream=sys.stdout,
        ).sort_stats(pstats.SortKey.CUMULATIVE)
        stats.print_stats()

    if pydm_args.faulthandler:
        faulthandler.disable()

    sys.exit(exit_code)


if __name__ == "__main__":
    main()



from pydm.about_pydm import AboutWindow


def test_about_window_launches(qtbot):
    """Make sure the About window doesn't crash."""
    a = AboutWindow(parent=None)
    qtbot.addWidget(a)
    a.show()



from ...utilities.remove_protocol import remove_protocol
from ...utilities.remove_protocol import protocol_and_address
from ...utilities.remove_protocol import parsed_address


def test_remove_protocol():
    out = remove_protocol("foo://bar")
    assert out == "bar"

    out = remove_protocol("bar")
    assert out == "bar"

    out = remove_protocol("foo://bar://foo2")
    assert out == "bar://foo2"


def test_protocol_and_address():
    out = protocol_and_address("foo://bar")
    assert out == ("foo", "bar")

    out = protocol_and_address("foo:/bar")
    assert out == (None, "foo:/bar")


def test_parsed_address():
    out = parsed_address(1)
    assert out is None

    out = parsed_address("foo:/bar")
    assert out is None

    out = parsed_address("foo://bar")
    assert out == ("foo", "bar", "", "", "", "")



import pytest
import logging
import numpy as np
from collections import OrderedDict
from ...widgets.channel import PyDMChannel
from ...widgets.scatterplot import ScatterPlotCurveItem, MINIMUM_BUFFER_SIZE, DEFAULT_BUFFER_SIZE
from ...utilities import remove_protocol

logger = logging.getLogger(__name__)


@pytest.mark.parametrize(
    "y_addr, x_addr, redraw_mode, bufferSizeChannelAddress, name",
    [
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, "test_name"),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, ""),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, None),
        ("", "", None, None, None),
        (None, None, None, None, None),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", ScatterPlotCurveItem.REDRAW_ON_BOTH, None, ""),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, "ca://test_value:Int", "test_name"),
    ],
)
def test_scatterplotcurveitem_construct(qtbot, y_addr, x_addr, redraw_mode, bufferSizeChannelAddress, name):
    plot_curve_item = ScatterPlotCurveItem(
        y_addr, x_addr, redraw_mode=redraw_mode, bufferSizeChannelAddress=bufferSizeChannelAddress, name=name
    )
    assert plot_curve_item is not None
    assert isinstance(plot_curve_item, ScatterPlotCurveItem)

    assert plot_curve_item._bufferSize == DEFAULT_BUFFER_SIZE
    assert plot_curve_item.redraw_mode == redraw_mode or ScatterPlotCurveItem.REDRAW_ON_EITHER
    assert np.array_equal(
        plot_curve_item.data_buffer, np.zeros((2, plot_curve_item._bufferSize), order="f", dtype=float)
    )
    for item in "x_connected y_connected bufferSizeChannel_connected".split():
        assert getattr(plot_curve_item, item) is False
    assert plot_curve_item.points_accumulated == 0
    for item in "latest_x_value latest_y_value".split():
        assert getattr(plot_curve_item, item) is None

    keys = (
        ("x_address", x_addr),
        ("y_address", y_addr),
        ("bufferSizeChannelAddress", bufferSizeChannelAddress),
    )
    for attr, kw in keys:
        obj = getattr(plot_curve_item, attr)
        expect = kw if kw else None
        assert obj == expect, f"{attr} {kw} {expect}"


@pytest.mark.parametrize(
    "y_addr, x_addr, redraw_mode, bufferSizeChannelAddress, name",
    [
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, "test_name"),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, ""),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, None, None),
        ("", "", None, None, None),
        (None, None, None, None, None),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", ScatterPlotCurveItem.REDRAW_ON_BOTH, None, ""),
        ("ca://test_value:FloatY", "ca://test_value:FloatX", None, "ca://test_value:Int", "test_name"),
    ],
)
def test_scatterplotcurveitem_to_dict(qtbot, y_addr, x_addr, redraw_mode, bufferSizeChannelAddress, name):
    plot_curve_item = ScatterPlotCurveItem(
        y_addr, x_addr, redraw_mode=redraw_mode, bufferSizeChannelAddress=bufferSizeChannelAddress, name=name
    )

    dictionary = plot_curve_item.to_dict()
    assert isinstance(dictionary, OrderedDict)

    y_name = remove_protocol(y_addr if y_addr is not None else "")
    x_name = remove_protocol(x_addr if x_addr is not None else "")
    if name is None:
        if y_addr is None and x_addr is None:
            assert dictionary["name"] == ""
        else:
            assert dictionary["name"] == f"{y_name} vs. {x_name}"
    else:
        assert dictionary["name"] == name


@pytest.mark.parametrize("new_address", ["new_address", "", None])
def test_scatterplotcurveitem_properties_and_setters(qtbot, new_address):
    plot_curve_item = ScatterPlotCurveItem(new_address, new_address, bufferSizeChannelAddress=new_address)

    assert plot_curve_item.x_address in (None, new_address)
    assert plot_curve_item.y_address in (None, new_address)
    assert plot_curve_item.bufferSizeChannelAddress in (None, new_address)

    if new_address:
        assert isinstance(plot_curve_item.x_channel, PyDMChannel)
        assert plot_curve_item.x_channel.address == new_address
        assert plot_curve_item.y_channel.address == new_address
    else:
        assert plot_curve_item.x_channel is None
        assert plot_curve_item.y_channel is None
        assert plot_curve_item.bufferSizeChannel is None


def test_scatterplotcurveitem_connection_state_changed(qtbot, signals):
    plot_curve_item = ScatterPlotCurveItem(None, None)
    assert plot_curve_item.x_connected is False
    assert plot_curve_item.y_connected is False

    signals.connection_state_signal.connect(plot_curve_item.xConnectionStateChanged)
    signals.connection_state_signal.emit(True)
    assert plot_curve_item.x_connected


@pytest.mark.parametrize(
    "redraw_mode, new_data",
    [
        (ScatterPlotCurveItem.REDRAW_ON_EITHER, [(0, 0), (1, 1), (2, 2), (2.5, 3.1)]),
        (ScatterPlotCurveItem.REDRAW_ON_BOTH, [(0, 0), (1, 1), (2, 2), (2.5, 3.1)]),
        (ScatterPlotCurveItem.REDRAW_ON_X, [(0, 0), (1, 1), (2, 2), (2.5, 3.1)]),
        (ScatterPlotCurveItem.REDRAW_ON_Y, [(0, 0), (1, 1), (2, 2), (2.5, 3.1)]),
    ],
)
def test_scatterplotcurveitem_receive_values(qtbot, signals, redraw_mode, new_data):
    # REDRAW_ON_X, REDRAW_ON_Y, REDRAW_ON_EITHER, REDRAW_ON_BOTH
    plot_curve_item = ScatterPlotCurveItem(None, None, redraw_mode=redraw_mode)

    expected_data_buffer = np.zeros((2, plot_curve_item._bufferSize), order="f", dtype=float)
    expected_data_buffer[0] = plot_curve_item.data_buffer[0]
    assert np.array_equal(expected_data_buffer, plot_curve_item.data_buffer)

    # inject the new_data, point by point
    assert plot_curve_item.points_accumulated == 0
    for i, pair in enumerate(new_data):
        new_x, new_y = pair

        plot_curve_item.receiveXValue(new_x)
        assert plot_curve_item.latest_x_value == new_x
        assert plot_curve_item.points_accumulated == 2 * i

        plot_curve_item.receiveYValue(new_y)
        assert plot_curve_item.latest_y_value == new_y
        assert plot_curve_item.points_accumulated == 2 * i + 1


def test_scatterplotcurve_initialize_buffer(qtbot):
    plot_curve_item = ScatterPlotCurveItem(None, None)

    plot_curve_item.initialize_buffer()

    assert plot_curve_item.points_accumulated == 0
    expected_data_buffer = np.zeros((2, plot_curve_item._bufferSize), order="f", dtype=float)
    expected_data_buffer[0] = plot_curve_item.data_buffer[0]

    assert np.array_equal(expected_data_buffer, plot_curve_item.data_buffer)


@pytest.mark.parametrize(
    "new_buffer_size, expected_set_buffer_size",
    [
        (0, MINIMUM_BUFFER_SIZE),
        (-5, MINIMUM_BUFFER_SIZE),
        (100, 100),
        (MINIMUM_BUFFER_SIZE + 1, MINIMUM_BUFFER_SIZE + 1),
    ],
)
def test_scatterplotcurve_get_set_reset_buffer_size(qtbot, new_buffer_size, expected_set_buffer_size):
    plot_curve_item = ScatterPlotCurveItem(None, None)

    assert plot_curve_item.getBufferSize() == DEFAULT_BUFFER_SIZE

    plot_curve_item.setBufferSize(new_buffer_size)
    assert plot_curve_item.getBufferSize() == expected_set_buffer_size

    plot_curve_item.resetBufferSize()
    assert plot_curve_item.getBufferSize() == DEFAULT_BUFFER_SIZE


@pytest.mark.parametrize(
    "addr, new_size, expected_buffer_size",
    [
        (None, None, DEFAULT_BUFFER_SIZE),
        ("", None, DEFAULT_BUFFER_SIZE),
        ("ca://test_value:Int", None, DEFAULT_BUFFER_SIZE),
    ],
)
def test_scatterplotcurve_get_set_reset_buffer_size_channel_addr(qtbot, addr, new_size, expected_buffer_size):
    plot_curve_item = ScatterPlotCurveItem(None, None, bufferSizeChannelAddress=addr)

    assert plot_curve_item.getBufferSize() == DEFAULT_BUFFER_SIZE

    if new_size is not None:
        plot_curve_item.bufferSizeChannelValueReceiver(new_size)
        assert plot_curve_item.getBufferSize() == new_size
        assert plot_curve_item.data_buffer.shape == (2, new_size)



from ..widgets.qtplugin_base import qtplugin_factory


def test_import_byte_plugin():
    # Byte plugin
    from ..widgets.byte import PyDMByteIndicator

    qtplugin_factory(PyDMByteIndicator)


def test_import_checkbox_plugin():
    # Checkbox plugin
    from ..widgets.checkbox import PyDMCheckbox

    qtplugin_factory(PyDMCheckbox)


def test_import_drawing_plugins():
    # Drawing plugins
    from ..widgets.drawing import (
        PyDMDrawingLine,
        PyDMDrawingRectangle,
        PyDMDrawingTriangle,
        PyDMDrawingEllipse,
        PyDMDrawingCircle,
        PyDMDrawingArc,
        PyDMDrawingPie,
        PyDMDrawingChord,
        PyDMDrawingImage,
        PyDMDrawingPolygon,
        PyDMDrawingPolyline,
        PyDMDrawingIrregularPolygon,
    )

    qtplugin_factory(PyDMDrawingImage)
    qtplugin_factory(PyDMDrawingLine)
    qtplugin_factory(PyDMDrawingRectangle)
    qtplugin_factory(PyDMDrawingTriangle)
    qtplugin_factory(PyDMDrawingEllipse)
    qtplugin_factory(PyDMDrawingCircle)
    qtplugin_factory(PyDMDrawingArc)
    qtplugin_factory(PyDMDrawingPie)
    qtplugin_factory(PyDMDrawingChord)
    qtplugin_factory(PyDMDrawingPolygon)
    qtplugin_factory(PyDMDrawingPolyline)
    qtplugin_factory(PyDMDrawingIrregularPolygon)


def test_import_embedded_display_plugin():
    # Embedded Display plugin
    from ..widgets.embedded_display import PyDMEmbeddedDisplay

    qtplugin_factory(PyDMEmbeddedDisplay)


def test_import_frame_plugin():
    # Frame plugin
    from ..widgets.frame import PyDMFrame

    qtplugin_factory(PyDMFrame, is_container=True)


def test_import_enum_button_plugin():
    # Enum Button plugin
    from ..widgets.enum_button import PyDMEnumButton

    qtplugin_factory(PyDMEnumButton)


def test_import_combobox_plugin():
    # Enum Combobox plugin
    from ..widgets.enum_combo_box import PyDMEnumComboBox

    qtplugin_factory(PyDMEnumComboBox)


def test_import_image_plugin():
    # Image plugin
    from ..widgets.image import PyDMImageView

    qtplugin_factory(PyDMImageView)


def test_import_label_plugin():
    # Label plugin
    from ..widgets.label import PyDMLabel

    qtplugin_factory(PyDMLabel)


def test_import_line_edit_plugin():
    # Line Edit plugin
    from ..widgets.line_edit import PyDMLineEdit

    qtplugin_factory(PyDMLineEdit)


def test_import_pushbutton_plugin():
    # Push Button plugin
    from ..widgets.pushbutton import PyDMPushButton

    qtplugin_factory(PyDMPushButton)


def test_import_related_display_plugin():
    # Related Display Button plugin
    from ..widgets.related_display_button import PyDMRelatedDisplayButton

    qtplugin_factory(PyDMRelatedDisplayButton)


def test_import_scale_indicator_plugin():
    # Scale Indicator plugin
    from ..widgets.scale import PyDMScaleIndicator

    qtplugin_factory(PyDMScaleIndicator)


def test_import_shellcmd_plugin():
    # Shell Command plugin
    from ..widgets.shell_command import PyDMShellCommand

    qtplugin_factory(PyDMShellCommand)


def test_import_slider_plugin():
    # Slider plugin
    from ..widgets.slider import PyDMSlider

    qtplugin_factory(PyDMSlider)


def test_import_spinbox_plugin():
    # Spinbox plugin
    from ..widgets.spinbox import PyDMSpinbox

    qtplugin_factory(PyDMSpinbox)


def test_import_symbol_plugin():
    # Symbol plugin
    from ..widgets.symbol import PyDMSymbol

    qtplugin_factory(PyDMSymbol)


def test_import_waveform_table_plugin():
    # Waveform Table plugin
    from ..widgets.waveformtable import PyDMWaveformTable

    qtplugin_factory(PyDMWaveformTable)


def test_import_timeplot_plugin():
    # Time Plot plugin
    from ..widgets.timeplot import PyDMTimePlot
    from ..widgets.timeplot_curve_editor import TimePlotCurveEditorDialog

    # Time Plot plugin
    qtplugin_factory(PyDMTimePlot, TimePlotCurveEditorDialog)


def test_import_waveformplot_plugin():
    # Time Plot plugin
    from ..widgets.waveformplot import PyDMWaveformPlot
    from ..widgets.waveformplot_curve_editor import WaveformPlotCurveEditorDialog

    # Waveform Plot plugin
    qtplugin_factory(PyDMWaveformPlot, WaveformPlotCurveEditorDialog)


def test_import_scatterplot_plugin():
    from ..widgets.scatterplot import PyDMScatterPlot
    from ..widgets.scatterplot_curve_editor import ScatterPlotCurveEditorDialog

    # Scatter Plot plugin
    qtplugin_factory(PyDMScatterPlot, ScatterPlotCurveEditorDialog)


def test_import_tab_widget_plugin():
    # Symbol plugin
    from ..widgets.tab_bar_qtplugin import TabWidgetPlugin

    TabWidgetPlugin()



from pyqtgraph.graphicsItems.ViewBox import ViewBox
from pyqtgraph.graphicsItems.ViewBox.ViewBoxMenu import ViewBoxMenu
from qtpy.QtCore import QCoreApplication, Signal
from qtpy.QtWidgets import QAction


class MultiAxisViewBoxMenu(ViewBoxMenu):
    """
    MultiAxisViewBoxMenu is a PyQtGraph ViewBoxMenu subclass to be used with MultiAxisViewBox. It can override
    the menu functionality to ensure that any selected option propagates through to each ViewBox in the plot.

    Parameters
    ----------
    view: ViewBox
        The view box this menu is associated with
    """

    # A signal indicating that the user has changed the mouse mode (left click panning vs. zooming)
    sigMouseModeChanged = Signal(object)
    # A signal indicating the user wants the default x and y axis ranges restored on this plot
    sigRestoreRanges = Signal()
    # A signal to set autorange for every view box on the plot
    sigSetAutorange = Signal(bool, bool)
    # A signal for updating the x autorange value
    sigXAutoRangeChanged = Signal(object)
    # A signal for updating the y autorange value
    sigYAutoRangeChanged = Signal(object)
    # A signal for inverting the x or y axis
    sigInvertAxis = Signal(int, bool)
    # Only panning when auto range is enabled (no scaling)
    sigAutoPan = Signal(object, object)
    # Auto range using only the visible portion of the plot when checked
    sigVisibleOnly = Signal(object, object)
    # Set the x range manually
    sigXManualRange = Signal(float, float)
    # Set the y range manually
    sigYManualRange = Signal(float, float)

    def __init__(self, view):
        super(MultiAxisViewBoxMenu, self).__init__(view)
        self.restoreRangesAction = QAction(QCoreApplication.translate("ViewBox", "Restore default X/Y ranges"), self)
        self.restoreRangesAction.triggered.connect(self.restoreRanges)

        # Insert/remove/insert because there is no QMenu function for inserting based on index
        self.insertAction(self.viewAll, self.restoreRangesAction)
        self.removeAction(self.viewAll)
        self.insertAction(self.restoreRangesAction, self.viewAll)

    def set3ButtonMode(self):
        """Change the mouse left-click functionality to pan the plot"""
        super(MultiAxisViewBoxMenu, self).set3ButtonMode()
        self.sigMouseModeChanged.emit("pan")

    def set1ButtonMode(self):
        """Change the mouse left-click functionality to zoom in on the plot"""
        super(MultiAxisViewBoxMenu, self).set1ButtonMode()
        self.sigMouseModeChanged.emit("rect")

    def xAutoClicked(self):
        """Update the auto-range value for each view box"""
        super().xAutoClicked()
        val = self.ctrl[0].autoPercentSpin.value() * 0.01
        self.sigXAutoRangeChanged.emit(val)

    def xManualClicked(self):
        """Disable x auto-range for each view box"""
        super().xManualClicked()
        self.sigXAutoRangeChanged.emit(False)

    def xRangeTextChanged(self):
        """Manually set the x-axis range to the user's input. Range will be unchanged if input was invalid"""
        super().xRangeTextChanged()
        updated_values = self._validateRangeText(ViewBox.XAxis)
        self.sigXManualRange.emit(*updated_values)

    def yAutoClicked(self):
        """Update the y auto-range value for each view box"""
        super().yAutoClicked()
        val = self.ctrl[1].autoPercentSpin.value() * 0.01
        self.sigYAutoRangeChanged.emit(val)

    def yManualClicked(self):
        """Disable y auto-range for each view box"""
        super().yManualClicked()
        self.sigYAutoRangeChanged.emit(False)

    def yRangeTextChanged(self):
        """Manually set the y-axis range to the user's input. Range will be unchanged if input was invalid"""
        super().yRangeTextChanged()
        updated_values = self._validateRangeText(ViewBox.YAxis)
        self.sigYManualRange.emit(*updated_values)

    def xAutoPanToggled(self, autoPan: bool):
        """Toggle the auto pan status of the x-axis"""
        super().xAutoPanToggled(autoPan)
        self.sigAutoPan.emit(autoPan, None)

    def xVisibleOnlyToggled(self, autoVisible: bool):
        """Toggle the visible only status of autorange for the x-axis"""
        super().xVisibleOnlyToggled(autoVisible)
        self.sigVisibleOnly.emit(autoVisible, None)

    def yAutoPanToggled(self, autoPan: bool):
        """Toggle the auto pan status of the y-axis"""
        super().yAutoPanToggled(autoPan)
        self.sigAutoPan.emit(None, autoPan)

    def yVisibleOnlyToggled(self, autoVisible: bool):
        """Toggle the visible only status of autorange for the y-axis"""
        super().yVisibleOnlyToggled(autoVisible)
        self.sigVisibleOnly.emit(None, autoVisible)

    def yInvertToggled(self, inverted: bool):
        """Toggle the inverted status of the y-axis."""
        super().yInvertToggled(inverted)
        self.sigInvertAxis.emit(ViewBox.YAxis, inverted)

    def xInvertToggled(self, inverted: bool):
        """Toggle the inverted status of the x-axis"""
        super().xInvertToggled(inverted)
        self.sigInvertAxis.emit(ViewBox.XAxis, inverted)

    def autoRange(self):
        """Sets autorange to True for all elements on the plot"""
        self.sigSetAutorange.emit(True, True)

    def restoreRanges(self):
        """Restore the original x and y axis ranges for this plot"""
        self.sigRestoreRanges.emit()



import numpy as np
from pyqtgraph import BarGraphItem
from unittest import mock
from ...widgets.waveformplot import PyDMWaveformPlot, WaveformCurveItem


@mock.patch("pydm.widgets.waveformplot.WaveformCurveItem.setData")
@mock.patch("pyqtgraph.BarGraphItem.setOpts")
def test_redraw_plot(mocked_set_opts, mocked_set_data, qtbot, monkeypatch):
    """Test redrawing a waveform plot using both a line and a bar graph"""

    # Create a waveform plot and add two data items to it, one to be rendered as a line and one as a bar graph
    waveform_plot = PyDMWaveformPlot()
    line_item = WaveformCurveItem()
    bar_item = WaveformCurveItem(plot_style="Bar")
    bar_item.bar_graph_item = BarGraphItem(x=[], height=[], width=1.0)
    waveform_plot.addCurve(line_item)
    waveform_plot.addCurve(bar_item)

    # Setup some mock data for our data items
    line_item.receiveXWaveform(np.array([1, 5, 10], dtype=float))
    line_item.receiveYWaveform(np.array([10, 15, 12], dtype=float))
    bar_item.receiveXWaveform(np.array([0.5, 1, 1.5, 2, 10, 11], dtype=float))
    bar_item.receiveYWaveform(np.array([45, 50, 52, 40, 24, 30], dtype=float))

    waveform_plot.set_needs_redraw()

    # Simulate a redraw of the plot
    waveform_plot.redrawPlot()

    # The line item should result in a call to set data displaying all available data points as defined above
    assert np.array_equal(mocked_set_data.call_args_list[2][1]["x"], np.array([1, 5, 10]))
    assert np.array_equal(mocked_set_data.call_args_list[2][1]["y"], np.array([10, 15, 12]))

    # As should the bar item, using the set_opts call instead of setData
    assert np.array_equal(mocked_set_opts.call_args_list[1][1]["x"], np.array([0.5, 1, 1.5, 2, 10, 11]))
    assert np.array_equal(mocked_set_opts.call_args_list[1][1]["height"], np.array([45, 50, 52, 40, 24, 30]))

    # After a call to redraw, the plot returns to this state until more data arrives
    assert not waveform_plot._needs_redraw


def test_mismatched_shapes(qtbot):
    """Test that the logic around waveforms with differing lengths works as expected"""

    # Create a waveform plot and add data whose waveform components do not share the same length
    PyDMWaveformPlot()
    data_item_1 = WaveformCurveItem()

    # Start with the basic case, both waveforms share the same length
    data_item_1.receiveXWaveform(np.array([1, 5, 10, 15], dtype=float))
    data_item_1.receiveYWaveform(np.array([10, 11, 12, 13], dtype=float))
    data_item_1.redrawCurve()

    # The data should remain exactly as is, no truncation required
    assert np.array_equal(data_item_1.x_waveform, np.array([1, 5, 10, 15]))
    assert np.array_equal(data_item_1.y_waveform, np.array([10, 11, 12, 13]))

    data_item_1.receiveXWaveform(np.array([1, 5, 10], dtype=float))
    data_item_1.receiveYWaveform(np.array([10, 11, 12, 13], dtype=float))
    data_item_1.redrawCurve()

    # Now the y-waveform was longer than the x one, so it gets truncated to match the length of x
    assert np.array_equal(data_item_1.x_waveform, np.array([1, 5, 10]))
    assert np.array_equal(data_item_1.y_waveform, np.array([10, 11, 12]))

    data_item_1.receiveXWaveform(np.array([1, 5, 10, 15], dtype=float))
    data_item_1.receiveYWaveform(np.array([10, 11, 12], dtype=float))
    data_item_1.redrawCurve()

    # Opposite case: the x-waveform was longer than the y one, so it gets truncated to match the length of y
    assert np.array_equal(data_item_1.x_waveform, np.array([1, 5, 10]))
    assert np.array_equal(data_item_1.y_waveform, np.array([10, 11, 12]))


def test_clear_curves(qtbot):
    """Verify that all curves are removed from a waveform plot when clearCurves() is called"""
    # Create a plot with two curves added to it
    waveform_plot = PyDMWaveformPlot()
    qtbot.addWidget(waveform_plot)
    data_one = WaveformCurveItem()
    data_two = WaveformCurveItem()
    waveform_plot.addCurve(data_one)
    waveform_plot.addCurve(data_two)

    assert len(waveform_plot.plotItem.curves) == 2  # Confirm that the curves were added properly

    waveform_plot.clearCurves()
    assert len(waveform_plot.plotItem.curves) == 0  # Now they should be gone


def test_clear_axes(qtbot):
    """Verify that when multiple y-axes are added to a plot, clearing out the curves and axes cleans up everything"""
    # Create a plot with two separate y axes on it, each with its own associated view box
    waveform_plot = PyDMWaveformPlot()
    qtbot.addWidget(waveform_plot)
    data_one = WaveformCurveItem()
    data_two = WaveformCurveItem()
    waveform_plot.addCurve(data_one, y_axis_name="Axis 1")
    waveform_plot.addCurve(data_two, y_axis_name="Axis 2")

    # Ensure both axes were properly added
    assert "Axis 1" in waveform_plot.plotItem.axes
    assert "Axis 2" in waveform_plot.plotItem.axes
    assert len(waveform_plot.plotItem.stackedViews) == 3  # There is a main top level view in addition to the 2 we added

    waveform_plot.clearAxes()

    # After the call to clear both axes should be removed, and the stacked views are also empty until more data is added
    assert "Axis 1" not in waveform_plot.plotItem.axes
    assert "Axis 2" not in waveform_plot.plotItem.axes
    assert len(waveform_plot.plotItem.stackedViews) == 0



from .qtplugin_base import PyDMDesignerPlugin, WidgetCategory
from .tab_bar import PyDMTabWidget


class TabWidgetPlugin(PyDMDesignerPlugin):
    """TabWidgetPlugin needs a custom plugin so that it can
    populate itself with an initial tab."""

    TabClass = PyDMTabWidget

    def __init__(self, extensions=None):
        super(TabWidgetPlugin, self).__init__(self.TabClass, group=WidgetCategory.CONTAINER, extensions=extensions)

    def domXml(self):
        """
        XML Description of the widget's properties.
        """
        return (
            '<widget class="{0}" name="{0}">\n'
            ' <property name="toolTip" >\n'
            "  <string>{1}</string>\n"
            " </property>\n"
            ' <property name="whatsThis" >\n'
            "  <string>{2}</string>\n"
            " </property>\n"
            '<property name="alarmChannels">\n'
            "<stringlist>\n"
            "  <string></string>\n"
            "</stringlist>\n"
            "</property>\n"
            '<widget class="QWidget" name="tab">\n'
            ' <attribute name="title">\n'
            "  <string>Page 1</string>\n"
            " </attribute>\n"
            "</widget>\n"
            "</widget>"
        ).format(self.name(), self.toolTip(), self.whatsThis())



from scipy import constants

UNITS = {
    "length": {
        "m": 1,
        "cm": constants.centi,
        "mm": constants.milli,
        "um": constants.micro,
        "nm": constants.nano,
        "pm": constants.pico,
        "in": constants.inch,
        "ft": constants.foot,
        "yds": constants.yard,
    },
    "time": {
        "s": 1,
        "ms": constants.milli,
        "us": constants.micro,
        "ns": constants.nano,
        "ps": constants.pico,
        "min": constants.minute,
        "hr": constants.hour,
        "weeks": constants.week,
        "days": constants.day,
    },
    "frequency": {
        "Hz": 1,
        "kHz": constants.kilo,
        "MHz": constants.mega,
        "GHz": constants.giga,
        "THz": constants.tera,
        "mHz": constants.milli,
    },
    "angle": {
        "rad": 1,
        "mrad": constants.milli,
        "urad": constants.micro,
        "nrad": constants.nano,
        "degree": constants.degree,
        "turn": 2 * constants.pi,
    },
    "voltage": {
        "V": 1,
        "MV": constants.mega,
        "kV": constants.kilo,
        "mV": constants.milli,
        "uV": constants.micro,
    },
    "current": {
        "A": 1,
        "MA": constants.mega,
        "kA": constants.kilo,
        "mA": constants.milli,
        "uA": constants.micro,
        "nA": constants.nano,
    },
}


def find_unittype(unit):
    """
    Find the type of a unit string.

    Parameters
    ----------
    unit : str
        The unit string

    Returns
    -------
    tp : str
        The unit type name or None if not found.
    """
    for tp in UNITS.keys():
        if unit in UNITS[tp].keys():
            return tp
    return None


def find_unit(unit):
    """
    Find the conversion of a unit string.

    Parameters
    ----------
    unit : str
        The unit string

    Returns
    -------
    float or None
        The unit value relative to the standard or None if not found.
    """
    tp = find_unittype(unit)
    if tp:
        return UNITS[tp][unit]
    else:
        return None


def convert(unit, desired):
    """
    Find the conversion rate of two different unit strings.

    Parameters
    ----------
    unit : str
        The current unit string
    desired : str
        The desired unit string

    Returns
    -------
    float or None
        The relation between unit and desired or None if not found.
    """
    current = find_unit(unit)
    final = find_unit(desired)

    if find_unittype(unit) != find_unittype(desired):
        return None

    if current and final:
        return current / final


def find_unit_options(unit):
    """
    Find the options for a given unit.

    Parameters
    ----------
    unit : str
        The unit string

    Returns
    -------
    list or None
        The list of similar units in crescent order or None if not found.
    """
    tp = find_unittype(unit)
    if tp:
        units = [choice for choice, _ in sorted(UNITS[tp].items(), key=lambda x: 1 / x[1])]
        return units
    else:
        return None



import os

from pydm import PyDMApplication
from pydm.display import Display, clear_compiled_ui_file_cache
from qtpy import uic
from unittest.mock import MagicMock, patch

# The path to the .ui file for creating a main window
test_ui_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), "test_data", "test.ui")


@patch("qtpy.uic.compileUi", wraps=uic.compileUi)
def test_reload_display(wrapped_compile_ui: MagicMock, qapp: PyDMApplication) -> None:
    """Verify that when a user reloads a PyDM window the underling display's ui file is actually reloaded"""
    clear_compiled_ui_file_cache()  # Ensure other tests have not already compiled our test file before we start

    try:
        display = Display(parent=None, ui_filename=test_ui_path)

        qapp.make_main_window()
        qapp.main_window.set_display_widget(display)

        # When the display is first created and loaded the underlying ui file gets compiled
        wrapped_compile_ui.assert_called_once()

        # Reloading should force a re-compile of the ui file to ensure any changes are picked up
        qapp.main_window.reload_display(True)
        assert wrapped_compile_ui.call_count == 2
    finally:
        clear_compiled_ui_file_cache()



from .application import PyDMApplication
from .display import Display
from .data_plugins import set_read_only
from .widgets import PyDMChannel
from ._version import get_versions

__all__ = [
    "PyDMApplication",
    "Display",
    "set_read_only",
    "PyDMChannel",
]

__version__ = get_versions()["version"]
del get_versions



import versioneer
from setuptools import setup, find_packages
import platform

# To use a consistent encoding
from codecs import open
from os import path, environ

cur_dir = path.abspath(path.dirname(__file__))

with open(path.join(cur_dir, "requirements.txt"), "r") as f:
    requirements = f.read().split()

dev_requirements = []

with open(path.join(cur_dir, "dev-requirements.txt"), "r") as f:
    dev_requirements = f.read().split()

with open(path.join(cur_dir, "README.md"), "r", encoding="utf-8") as f:
    long_description = f.read()

# Remove the 'optional' requirements
optional = ("PyQt5", "PySide", "psutil", "pcaspy", "pyepics")
for package in optional:
    if package in requirements:
        requirements.remove(package)

extras_require = {"PySide": ["PySide"], "pyepics": ["pyepics"], "perf": ["psutil"], "dev": dev_requirements}

if "CONDA_PREFIX" not in environ:
    extras_require["PyQt5"] = ["PyQt5"]
else:
    print("******************************************************************")
    print("*                              WARNING                           *")
    print("******************************************************************")
    print("Installing at an Anaconda Environment, to avoid naming conflicts ")
    print("make sure you do:")
    print("conda install pyqt=5")
    print("******************************************************************")
    print("For more info please check: ")
    print("https://github.com/ContinuumIO/anaconda-issues/issues/1554")
    print("******************************************************************")

extras_require["all"] = sorted(set(sum(extras_require.values(), [])))
# Preference for PyQt5 if you select ALL...
extras_require["all"].remove("PySide")

entry_style = "gui_scripts"
if platform.system() == "Windows":
    entry_style = "console_scripts"

setup(
    name="pydm",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    # Author details
    author="SLAC National Accelerator Laboratory",
    packages=find_packages(),
    package_dir={"pydm": "pydm", "pydm_launcher": "pydm_launcher"},
    description="Python Display Manager",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/slaclab/pydm",
    entry_points={entry_style: ["pydm=pydm_launcher.main:main"]},
    license="BSD",
    install_requires=requirements,
    extras_require=extras_require,
    include_package_data=True,
    classifiers=[
        "License :: OSI Approved :: BSD License",
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
)



__all__ = [
    "PyDMChannel",
    "PyDMByteIndicator",
    "PyDMCheckbox",
    "PyDMDrawingLine",
    "PyDMDrawingRectangle",
    "PyDMDrawingTriangle",
    "PyDMDrawingEllipse",
    "PyDMDrawingCircle",
    "PyDMDrawingArc",
    "PyDMDrawingPie",
    "PyDMDrawingChord",
    "PyDMDrawingImage",
    "PyDMDrawingPolyline",
    "PyDMDrawingPolygon",
    "PyDMDrawingIrregularPolygon",
    "PyDMEmbeddedDisplay",
    "PyDMEnumComboBox",
    "PyDMImageView",
    "PyDMLabel",
    "PyDMLineEdit",
    "PyDMPushButton",
    "PyDMRelatedDisplayButton",
    "PyDMShellCommand",
    "PyDMSlider",
    "PyDMSpinbox",
    "PyDMSymbol",
    "PyDMWaveformTable",
    "PyDMScaleIndicator",
    "PyDMTimePlot",
    "PyDMArchiverTimePlot",
    "PyDMWaveformPlot",
    "PyDMScatterPlot",
    "PyDMEventPlot",
    "PyDMTabWidget",
    "PyDMTemplateRepeater",
    "PyDMNTTable",
]

from .channel import PyDMChannel
from .byte import PyDMByteIndicator
from .checkbox import PyDMCheckbox
from .drawing import (
    PyDMDrawingLine,
    PyDMDrawingRectangle,
    PyDMDrawingTriangle,
    PyDMDrawingEllipse,
    PyDMDrawingCircle,
    PyDMDrawingArc,
    PyDMDrawingPie,
    PyDMDrawingChord,
    PyDMDrawingImage,
    PyDMDrawingPolyline,
    PyDMDrawingPolygon,
    PyDMDrawingIrregularPolygon,
)
from .embedded_display import PyDMEmbeddedDisplay
from .enum_combo_box import PyDMEnumComboBox
from .image import PyDMImageView
from .label import PyDMLabel
from .line_edit import PyDMLineEdit
from .pushbutton import PyDMPushButton
from .related_display_button import PyDMRelatedDisplayButton
from .shell_command import PyDMShellCommand
from .slider import PyDMSlider
from .spinbox import PyDMSpinbox
from .symbol import PyDMSymbol
from .waveformtable import PyDMWaveformTable
from .scale import PyDMScaleIndicator
from .timeplot import PyDMTimePlot
from .archiver_time_plot import PyDMArchiverTimePlot
from .waveformplot import PyDMWaveformPlot
from .scatterplot import PyDMScatterPlot
from .eventplot import PyDMEventPlot
from .tab_bar import PyDMTabWidget
from .template_repeater import PyDMTemplateRepeater
from .nt_table import PyDMNTTable



import logging
import numpy as np
import collections
import threading
import p4p
import re
import time
from p4p.client.thread import Context, Disconnected
from p4p.wrapper import Value
from p4p.nt import NTURI
from .pva_codec import decompress
from pydm.data_plugins import is_read_only
from pydm.data_plugins.plugin import PyDMPlugin, PyDMConnection
from pydm.widgets.channel import PyDMChannel
from qtpy.QtCore import QObject, Qt
from typing import Optional
from urllib.parse import urlparse, parse_qs

logger = logging.getLogger(__name__)

# arbitrary default for non-polled RPC
DEFAULT_RPC_TIMEOUT = 5.0


class Connection(PyDMConnection):
    def __init__(
        self, channel: PyDMChannel, address: str, protocol: Optional[str] = None, parent: Optional[QObject] = None
    ):
        """
        Manages the connection to a channel using the P4P library. A given channel can have multiple listeners.
        Parameters
        ----------
        channel : PyDMChannel
            The channel that this connection is connected to.
        address : str
             The address of the PV.
        protocol : str, optional
             The protocol prepended to the address.
        parent : QObject, optional
             The parent object of this connection.
        """
        super().__init__(channel, address, protocol, parent)
        self._connected = False
        self.nttable_data_location = PyDMPlugin.get_subfield(channel)
        self._value = None
        self._severity = None
        self._precision = None
        self._enum_strs = None
        self._units = None
        self._upper_ctrl_limit = None
        self._lower_ctrl_limit = None
        self._upper_alarm_limit = None
        self._lower_alarm_limit = None
        self._upper_warning_limit = None
        self._lower_warning_limit = None
        self._timestamp = None

        # RPC = Remote Procedure Call (https://mdavidsaver.github.io/p4p/rpc.html#p4p.rpc.rpcproxy)
        # example address: pva://pv:call:add?lhs=4&rhs=7&pydm_pollrate=10
        self._rpc_function_name = ""  # pv:call:add (in case of above example)
        self._rpc_arg_names = []  # ['lhs', 'rhs'] (in case of above example)
        self._rpc_arg_values = []  # ['4', '7'] (in case of above example)
        self._value_obj = None
        # Poll rate in seconds
        self._rpc_poll_rate = 0  # (in case of above example)
        self._background_polling_thread = None

        self.monitor = None
        self.is_rpc = self.is_rpc_address(channel.address)
        if self.is_rpc:
            # channel.address provides the entire user-entered channel (instead of 'channel' var)
            self.parse_rpc_channel(channel.address)

        # RPC requests are handled simply and don't require continuous monitoring,
        # instead they use the p4p 'rpc' call at a specified a pollrate.
        self.add_listener(channel)
        if not self.is_rpc:
            self.monitor = P4PPlugin.context.monitor(name=self.address, cb=self.send_new_value, notify_disconnect=True)

    def emit_for_type(self, value) -> None:
        # Emit for the types currently supported as RPC request args
        if isinstance(value, int):
            self.new_value_signal[int].emit(value)
        elif isinstance(value, float):
            self.new_value_signal[float].emit(value)
        elif isinstance(value, bool):
            self.new_value_signal[bool].emit(value)
        elif isinstance(value, str):
            self.new_value_signal[str].emit(value)

    def poll_rpc_channel(self) -> None:
        # Keep executing this function at the polling rate

        # When polling-rate is not specified by user (is 0), just do a single RPC request
        only_poll_once = False
        if self._rpc_poll_rate == 0:
            self._rpc_poll_rate = DEFAULT_RPC_TIMEOUT
            only_poll_once = True

        while True:
            start_time = time.process_time()

            result = None
            try:
                result = P4PPlugin.context.rpc(
                    name=self._rpc_function_name, value=self._value_obj, timeout=self._rpc_poll_rate
                )
            except Exception:
                # So widget displays name of channel when can't connect to RPC channel
                self.connection_state_signal.emit(False)

            if result:
                self.connection_state_signal.emit(True)
                self.emit_for_type(result.value)
            else:
                self.connection_state_signal.emit(False)

            if only_poll_once:
                break

            rpc_call_time = time.process_time() - start_time
            # We want to call "rpc" every self._rpc_poll_rate seconds,
            # so wait when the call returns faster than the polling-rate.
            # The timeout arg makes sure a single call is never slower then the polling-rate.
            poll_rate_and_rpc_call_time_dif = self._rpc_poll_rate - rpc_call_time
            if poll_rate_and_rpc_call_time_dif > 0:
                time.sleep(poll_rate_and_rpc_call_time_dif)

    def get_arg_datatype(self, arg_value_string):
        # Try to figure out the datatype of RPC request args
        try:
            int(arg_value_string)
            return "i", int(arg_value_string)
        except Exception:
            pass
        try:
            float(arg_value_string)
            return "f", float(arg_value_string)
        except Exception:
            pass
        if arg_value_string.lower() == "True" or arg_value_string.lower() == "False":
            return "?", bool(arg_value_string)
        # Assume arg is just a string if no other type works
        return "s", arg_value_string

    def create_request(self, rpc_function_name, rpc_arg_names, rpc_arg_values) -> Value:
        # example addr: pv:call:add_two_ints?a=2&b=7&
        arg_datatypes = []
        for i in range(len(rpc_arg_names)):
            data_type, _ = self.get_arg_datatype(rpc_arg_values[i])
            if data_type is None:
                return None
            arg_datatypes.append((rpc_arg_names[i], data_type))
        # example arg_datatypes: [('a', 'i'), ('b', 'i')]
        arg_val_mapping = {key: value for (key, _), value in zip(arg_datatypes, rpc_arg_values)}
        # example arg_val_mapping: {'a': '2', 'b': '7'}

        # https://mdavidsaver.github.io/p4p/nt.html#p4p.nt.NTURI
        nturi_obj = NTURI(arg_datatypes)

        request = nturi_obj.wrap(rpc_function_name, scheme="pva", kws=arg_val_mapping)
        return request

    def parse_rpc_channel(self, input_string) -> None:
        # url parsing is close to what we need, so use with some adjusting
        parsed_url = urlparse(input_string)
        raw_args = parsed_url.query
        parsed_args = parse_qs(raw_args)
        function_name = parsed_url.netloc

        # if RPC has no args and no polling specified, url parsing will leave name with ending '&' char we need remove

        if len(function_name) >= 1 and function_name[-1] == "&":
            function_name = function_name[:-1]

        # now handle case when no args givin but specified polling
        pollrate = 0.0
        if "pydm_pollrate" in function_name:
            index = function_name.find("&pydm_pollrate=")
            if index != -1:
                value_str = function_name[index + len("&pydm_pollrate=") :]
                pollrate = value_str

            function_name = function_name.split("&")[0]
        else:
            # because url-parsing function put value-string in 1 item list
            pollrate = parsed_args.get("pydm_pollrate", "0.0")[0]
            if "pydm_pollrate" in parsed_args:
                # delete because we don't pass pollrate as argument to RPC
                del parsed_args["pydm_pollrate"]

        for curr_arg_name, curr_arg_value in parsed_args.items():
            parsed_args[curr_arg_name] = curr_arg_value[0]  # [0] takes value out of 1 item list

        self._rpc_function_name = function_name
        self._rpc_arg_names = list(parsed_args.keys())
        self._rpc_arg_values = list(parsed_args.values())
        self._rpc_poll_rate = float(pollrate)

    def is_rpc_address(self, full_channel_name):
        """
        Keep this simple for now, say it's an RPC just if either ends with '&' or '&pydm_pollrate=<number>.
        This should be enough to differentiate between non-rpc requests,
        bad RPCs will just fail and log error when we try to connect latter.
        """
        if full_channel_name is None:
            return False
        pattern = re.compile(r"(&|\&pydm_pollrate=\d+(\.\d+)?)$")
        return bool(pattern.search(full_channel_name))

    def clear_cache(self) -> None:
        """Clear out all the stored values of this connection."""
        self._value = None
        self._severity = None
        self._precision = None
        self._enum_strs = None
        self._units = None
        self._upper_ctrl_limit = None
        self._lower_ctrl_limit = None
        self._upper_alarm_limit = None
        self._lower_alarm_limit = None
        self._upper_warning_limit = None
        self._lower_warning_limit = None
        self._timestamp = None

    def send_new_value(self, value: Value) -> None:
        """Callback invoked whenever a new value is received by our monitor. Emits signals based on values changed."""
        if isinstance(value, Disconnected):
            self._connected = False
            self.clear_cache()
            self.connection_state_signal.emit(False)
        else:
            if not self._connected:
                self._connected = True
                self.connection_state_signal.emit(True)
                # Note that there is no way to get the actual write access value from p4p, so defaulting to True for now
                self.write_access_signal.emit(True)

            self._value = value
            has_value_changed_yet = False
            for changed_value in value.changedSet():
                if changed_value == "value" or changed_value.split(".")[0] == "value":
                    # NTTable has a changedSet item for each column that has changed
                    # Since we want to send an update on any table change, let's track
                    # if the value item has been updated yet
                    if has_value_changed_yet:
                        continue
                    else:
                        has_value_changed_yet = True

                    if "NTTable" in value.getID():
                        new_value = value.value.todict()
                        if hasattr(value, "labels") and "labels" not in new_value:
                            # Labels are the column headers for the table
                            new_value["labels"] = value.labels
                    elif "NTEnum" in value.getID():
                        new_value = value.value.index
                        self.enum_strings_signal.emit(tuple(value.value.choices))
                    else:
                        new_value = value.value

                    if self.nttable_data_location:
                        msg = f"Invalid channel... {self.nttable_data_location}"
                        for subfield in self.nttable_data_location:
                            if isinstance(new_value, collections.abc.Container) and not isinstance(new_value, str):
                                if isinstance(subfield, str):
                                    try:
                                        new_value = new_value[subfield]
                                        continue
                                    except (TypeError, IndexError):
                                        logger.debug(
                                            """Type Error when attempting to use the given key, code will next attempt
                                            to convert the key to an int"""
                                        )
                                    except KeyError:
                                        logger.exception(msg)

                                    try:
                                        new_value = new_value[int(subfield)]
                                    except ValueError:
                                        logger.exception(msg, exc_info=True)
                            else:
                                logger.exception(msg, exc_info=True)
                                raise ValueError(msg)

                    if new_value is not None:
                        if isinstance(new_value, np.ndarray):
                            if "NTNDArray" in value.getID():
                                new_value = decompress(value)
                            self.new_value_signal[np.ndarray].emit(new_value)
                        elif isinstance(new_value, np.bool_):
                            self.new_value_signal[np.bool_].emit(new_value)
                        elif isinstance(new_value, list):
                            self.new_value_signal[np.ndarray].emit(np.array(new_value))
                        elif isinstance(new_value, float):
                            self.new_value_signal[float].emit(new_value)
                        elif isinstance(new_value, int):
                            self.new_value_signal[int].emit(new_value)
                        elif isinstance(new_value, str):
                            self.new_value_signal[str].emit(new_value)
                        elif isinstance(new_value, dict):
                            self.new_value_signal[dict].emit(new_value)
                        else:
                            raise ValueError(f"No matching signal for value: {new_value} with type: {type(new_value)}")
                # Sometimes unchanged control variables appear to be returned with value changes, so checking against
                # stored values to avoid sending misleading signals. Will revisit on data plugin changes.
                elif changed_value == "alarm.severity" and value.alarm.severity != self._severity:
                    self._severity = value.alarm.severity
                    self.new_severity_signal.emit(value.alarm.severity)
                elif changed_value == "display.precision" and value.display.precision != self._precision:
                    self._precision = value.display.precision
                    self.prec_signal.emit(value.display.precision)
                elif changed_value == "display.units" and value.display.units != self._units:
                    self._units = value.display.units
                    self.unit_signal.emit(value.display.units)
                elif changed_value == "control.limitLow" and value.control.limitLow != self._lower_ctrl_limit:
                    self._lower_ctrl_limit = value.control.limitLow
                    self.lower_ctrl_limit_signal.emit(value.control.limitLow)
                elif changed_value == "control.limitHigh" and value.control.limitHigh != self._upper_ctrl_limit:
                    self._upper_ctrl_limit = value.control.limitHigh
                    self.upper_ctrl_limit_signal.emit(value.control.limitHigh)
                elif (
                    changed_value == "valueAlarm.highAlarmLimit"
                    and value.valueAlarm.highAlarmLimit != self._upper_alarm_limit
                ):
                    self._upper_alarm_limit = value.valueAlarm.highAlarmLimit
                    self.upper_alarm_limit_signal.emit(value.valueAlarm.highAlarmLimit)
                elif (
                    changed_value == "valueAlarm.lowAlarmLimit"
                    and value.valueAlarm.lowAlarmLimit != self._lower_alarm_limit
                ):
                    self._lower_alarm_limit = value.valueAlarm.lowAlarmLimit
                    self.lower_alarm_limit_signal.emit(value.valueAlarm.lowAlarmLimit)
                elif (
                    changed_value == "valueAlarm.highWarningLimit"
                    and value.valueAlarm.highWarningLimit != self._upper_warning_limit
                ):
                    self._upper_warning_limit = value.valueAlarm.highWarningLimit
                    self.upper_warning_limit_signal.emit(value.valueAlarm.highWarningLimit)
                elif (
                    changed_value == "valueAlarm.lowWarningLimit"
                    and value.valueAlarm.lowWarningLimit != self._lower_warning_limit
                ):
                    self._lower_warning_limit = value.valueAlarm.lowWarningLimit
                    self.lower_warning_limit_signal.emit(value.valueAlarm.lowWarningLimit)
                elif (
                    changed_value == "timeStamp.secondsPastEpoch"
                    and value.timeStamp.secondsPastEpoch != self._timestamp
                ):
                    self._timestamp = value.timeStamp.secondsPastEpoch
                    self.timestamp_signal.emit(value.timeStamp.secondsPastEpoch)

    @staticmethod
    def convert_epics_nttable(epics_struct):
        """
        Converts an epics nttable (passed as a class object p4p.wrapper.Value) to a python dictionary.

        Parameters
        ----------
        epics_struct: 'p4p.wrapper.Value'

        Return
        ------
        result: dict
        """
        result = {}
        for field in epics_struct.keys():
            value = epics_struct[field]
            if isinstance(value, np.ndarray):
                value = value.tolist()
            elif isinstance(value, p4p.wrapper.Value):
                value = Connection.convert_epics_nttable(value)
            result[field] = value
        return result

    @staticmethod
    def set_value_by_keys(table, keys, new_value):
        """
        Saves the passed new_value into the appropriate spot in the given table
        using the given keys.

        Parameters
        ----------
        table: dict
        keys: list
        new_value: str
        """
        if len(keys) == 1:
            key = keys[0]
            try:
                table[key] = new_value
            except TypeError:
                table[int(key)] = new_value
        else:
            key = keys[0]

            Connection.set_value_by_keys(table[key], keys[1:], new_value)

    def put_value(self, value):
        """Write a value to the PV"""

        if self.nttable_data_location:
            nttable = Connection.convert_epics_nttable(self._value)
            nttable = nttable["value"]
            Connection.set_value_by_keys(nttable, self.nttable_data_location, value)
            value = {"value": nttable}
        if is_read_only():
            logger.warning(f"PyDM read-only mode is enabled, could not write value: {value} to {self.address}")
            return

        if self.is_rpc:
            return
        try:
            P4PPlugin.context.put(self.monitor.name, value)
        except Exception as e:
            logger.error(f"Unable to put value: {value} to channel {self.monitor.name}: {e}")

    def add_listener(self, channel: PyDMChannel):
        """
        Adds a listener to this connection, connecting the appropriate signals/slots to the input PyDMChannel.
        Parameters
        ----------
        channel : PyDMChannel
            The channel that will be listening to any changes from this connection
        """
        super().add_listener(channel)

        if self.is_rpc:
            # In case of a RPC, we can just query the channel immediately and emit the value,
            # and let the pollrate dictate if/when we query and emit again.f
            self._value_obj = self.create_request(self._rpc_function_name, self._rpc_arg_names, self._rpc_arg_values)
            if self._value_obj is None:
                logger.warning(f"failed to create request object for RPC to {self._rpc_function_name}")
                return

            # Use daemon threads so they will be stopped when all the non-daemon
            # threads (in our case just the main thread) are killed, preventing them from running forever.
            self._background_polling_thread = threading.Thread(target=self.poll_rpc_channel, daemon=True)
            self._background_polling_thread.start()

            return

        if self.monitor is not None and self._connected:
            # Adding a listener to an already connected PV. Manually send the signals indicating the PV is
            # connected, and what the last known values were.
            self.connection_state_signal.emit(True)
            self.write_access_signal.emit(True)
            value_to_send = self._value
            self.clear_cache()
            if value_to_send is not None:
                self.send_new_value(value_to_send)

        if channel.value_signal is not None:
            try:
                channel.value_signal[str].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[int].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[float].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[np.ndarray].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass
            try:
                channel.value_signal[dict].connect(self.put_value, Qt.QueuedConnection)
            except KeyError:
                pass

    def close(self):
        """Closes out this connection."""
        # If RPC, we have no monitor to close
        if self.monitor:
            self.monitor.close()
        super().close()


class P4PPlugin(PyDMPlugin):
    # NOTE: protocol is intentionally "None" to keep this plugin from getting directly imported.
    # If this plugin is chosen as the One True PVA Plugin in pva_plugin.py, the protocol will
    # be properly set before it is used.
    protocol = None
    connection_class = Connection
    context = None

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if P4PPlugin.context is None:
            # Create the p4p pva context for all connections to use
            context = Context("pva", nt=False)  # Disable automatic value unwrapping
            P4PPlugin.context = context



import os
import json
import copy
import pytest
import webbrowser

from qtpy import QtCore
from qtpy.QtWidgets import QMessageBox, QTableWidgetSelectionRange

from ...widgets.base import PyDMPrimitiveWidget
from ...widgets.rules_editor import RulesEditor


class DummyWidget:
    """
    A stub class to play with the rules editor and not touch
    the RulesEngine
    """

    RULE_PROPERTIES = PyDMPrimitiveWidget.RULE_PROPERTIES
    DEFAULT_RULE_PROPERTY = PyDMPrimitiveWidget.DEFAULT_RULE_PROPERTY

    def __init__(self, *args, **kwargs):
        self._rules = None

    @property
    def rules(self):
        return self._rules

    @rules.setter
    def rules(self, new_rules):
        if new_rules != self._rules:
            self._rules = new_rules


@pytest.mark.parametrize("use_enum, visible", [(None, False), (True, True), (False, False)])
def test_rules_editor(use_enum, visible, qtbot, monkeypatch):
    """
    Test the rules editor in general.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    monkeypatch : fixture
        To override dialog behaviors
    """

    # Create the base widget
    widget = DummyWidget()

    # Ensure that no rules are set
    assert widget.rules is None

    # Create a rule editor for this widget
    empty = RulesEditor(widget)
    qtbot.addWidget(empty)

    empty.show()

    qtbot.waitExposed(empty, timeout=5000)

    # Abort the changes
    empty.cancelChanges()

    # Create the rules data for the widget
    rules_list = [
        {
            "name": "Rule #1",
            "property": "Enable",
            "initial_value": "False",
            "expression": "ch[0] > 1",
            "channels": [{"channel": "ca://MTEST:Float", "trigger": True}],
        }
    ]
    if use_enum is not None:
        rules_list[0]["channels"][0]["use_enum"] = use_enum

    # Add the rules to the widget
    widget.rules = json.dumps(rules_list)

    # Create a new Editor Window
    re = RulesEditor(widget)
    qtbot.addWidget(re)
    re.show()

    qtbot.waitExposed(re, timeout=5000)

    assert re.lst_rules.count() == 1
    assert not re.frm_edit.isEnabled()

    ch_choices = {True: QtCore.Qt.Checked, False: QtCore.Qt.Unchecked}

    re.lst_rules.setCurrentRow(0)
    re.load_from_list()
    assert re.frm_edit.isEnabled()
    assert re.txt_name.text() == "Rule #1"
    assert re.cmb_property.currentText() == "Enable"
    assert re.tbl_channels.rowCount() == 1
    assert re.tbl_channels.item(0, 0).text() == "ca://MTEST:Float"
    assert re.tbl_channels.item(0, 1).checkState() == QtCore.Qt.Checked
    assert re.tbl_channels.item(0, 2).checkState() == ch_choices[visible]
    assert re.lbl_expected_type.text() == "bool"
    assert re.txt_expression.text() == "ch[0] > 1"
    assert re.txt_initial_value.text() == "False"

    qtbot.keyClicks(re.txt_name, "-Test")
    qtbot.keyClick(re.txt_name, QtCore.Qt.Key_Return)
    assert re.txt_name.text() == "Rule #1-Test"
    assert re.rules[0]["name"] == "Rule #1-Test"

    qtbot.mouseClick(re.btn_add_channel, QtCore.Qt.LeftButton)
    re.tbl_channels.item(1, 0).setText("ca://TEST")
    assert re.rules[0]["channels"][1]["channel"] == "ca://TEST"
    assert re.rules[0]["channels"][1]["trigger"] is False
    assert re.rules[0]["channels"][1]["use_enum"] is True

    re.txt_expression.clear()
    qtbot.keyClicks(re.txt_expression, "ch[0] < 1")
    qtbot.keyClick(re.txt_expression, QtCore.Qt.Key_Return)
    assert re.txt_expression.text() == "ch[0] < 1"
    assert re.rules[0]["expression"] == "ch[0] < 1"

    re.txt_initial_value.clear()
    qtbot.keyClicks(re.txt_initial_value, "True")
    qtbot.keyClick(re.txt_initial_value, QtCore.Qt.Key_Return)
    assert re.txt_initial_value.text() == "True"
    assert re.rules[0]["initial_value"] == "True"

    # Test Delete with Confirm - NO
    assert re.tbl_channels.rowCount() == 2
    re.tbl_channels.setRangeSelected(QTableWidgetSelectionRange(1, 0, 1, 1), True)
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.No)
    qtbot.mouseClick(re.btn_del_channel, QtCore.Qt.LeftButton)
    assert re.tbl_channels.rowCount() == 2

    # Test Delete with Confirm - YES
    re.tbl_channels.setRangeSelected(QTableWidgetSelectionRange(1, 0, 1, 1), True)
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.Yes)
    qtbot.mouseClick(re.btn_del_channel, QtCore.Qt.LeftButton)
    assert re.tbl_channels.rowCount() == 1
    assert len(re.rules[0]["channels"]) == 1

    # Test Delete with Invalid Selection
    re.tbl_channels.setRangeSelected(QTableWidgetSelectionRange(1, 0, 1, 1), True)
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.Yes)
    qtbot.mouseClick(re.btn_del_channel, QtCore.Qt.LeftButton)
    assert re.tbl_channels.rowCount() == 1
    assert len(re.rules[0]["channels"]) == 1

    qtbot.mouseClick(re.btn_add_rule, QtCore.Qt.LeftButton)
    assert re.lst_rules.count() == 2
    assert re.frm_edit.isEnabled()
    assert re.txt_name.text() == "New Rule"
    assert re.cmb_property.currentText() == widget.DEFAULT_RULE_PROPERTY
    assert re.tbl_channels.rowCount() == 0
    assert re.txt_expression.text() == ""

    qtbot.mouseClick(re.btn_add_channel, QtCore.Qt.LeftButton)
    assert re.tbl_channels.item(0, 0).text() == ""
    assert re.tbl_channels.item(0, 1).checkState() == QtCore.Qt.Checked
    assert re.tbl_channels.item(0, 2).checkState() == QtCore.Qt.Checked

    qtbot.mouseClick(re.btn_add_channel, QtCore.Qt.LeftButton)
    assert re.tbl_channels.item(1, 0).text() == ""
    assert re.tbl_channels.item(1, 1).checkState() == QtCore.Qt.Unchecked
    assert re.tbl_channels.item(1, 2).checkState() == QtCore.Qt.Checked

    # Switch between the rules
    re.lst_rules.setCurrentRow(0)
    re.lst_rules.setCurrentRow(1)

    # Delete Rule 1 - Confirm - NO
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.No)
    qtbot.mouseClick(re.btn_del_rule, QtCore.Qt.LeftButton)
    assert re.lst_rules.count() == 2

    # Delete Rule 1 - Confirm - YES
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.Yes)
    qtbot.mouseClick(re.btn_del_rule, QtCore.Qt.LeftButton)
    assert re.frm_edit.isEnabled() is False
    assert re.lst_rules.count() == 1

    re.lst_rules.setCurrentRow(0)
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.Yes)
    qtbot.mouseClick(re.btn_del_rule, QtCore.Qt.LeftButton)
    assert re.frm_edit.isEnabled() is False
    assert re.lst_rules.count() == 0

    # Delete Empty List - Confirm - YES
    monkeypatch.setattr(QMessageBox, "question", lambda *args: QMessageBox.Yes)
    qtbot.mouseClick(re.btn_del_rule, QtCore.Qt.LeftButton)
    assert re.frm_edit.isEnabled() is False
    assert re.lst_rules.count() == 0
    re.cancelChanges()


def test_rules_editor_data_valid(qtbot):
    """
    Test the rules form validation.

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    """

    def validate(expected_status, expected_msg, definition):
        status, msg = RulesEditor.is_data_valid(definition)
        assert status == expected_status
        assert expected_msg in msg

    rules_list = [
        {
            "name": "Rule #1",
            "property": "Enable",
            "expression": "ch[0] > 1",
            "channels": [{"channel": "ca://MTEST:Float", "trigger": True}],
        }
    ]

    validate(True, "", rules_list)

    rules_original = copy.deepcopy(rules_list)

    rules_original[0]["name"] = ""
    validate(False, "has no name", rules_original)

    rules_original = copy.deepcopy(rules_list)
    rules_original[0]["expression"] = ""
    validate(False, "has no expression", rules_original)

    rules_original = copy.deepcopy(rules_list)
    old_channels = rules_original[0]["channels"]
    rules_original[0]["channels"] = []
    validate(False, "has no channel", rules_original)
    rules_original[0]["channels"] = old_channels

    rules_original[0]["channels"][0]["trigger"] = False
    validate(False, "has no channel for trigger", rules_original)

    rules_original[0]["channels"][0]["channel"] = None
    validate(False, "Ch. #0 has no channel.", rules_original)
    rules_original[0]["channels"][0]["channel"] = ""
    validate(False, "Ch. #0 has no channel.", rules_original)


def test_rules_editor_open_help(qtbot, monkeypatch):
    """
    Test the Open Help button

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    monkeypatch : fixture
        To override dialog behaviors
    """
    # Create the base widget
    widget = DummyWidget()

    # Create a new Editor Window
    re = RulesEditor(widget)
    qtbot.addWidget(re)
    re.show()

    qtbot.waitExposed(re, timeout=5000)

    re.lst_rules.setCurrentRow(0)

    url = re.open_help(open=False)
    base_url = os.getenv("PYDM_DOCS_URL", "https://slaclab.github.io/pydm")
    exp_url = base_url + "/widgets/widget_rules/index.html"
    assert url == exp_url

    monkeypatch.setattr(webbrowser, "open", lambda *args, **kwargs: "")
    re.open_help()
    re.cancelChanges()


def test_rules_editor_add_notes(qtbot, monkeypatch):
    """
    Test adding notes for a widget's rules

    Parameters
    ----------
    qtbot : fixture
        pytest-qt window for widget test
    monkeypatch : fixture
        To override dialog behaviors
    """
    # Create the base widget
    widget = DummyWidget()

    # Create the rules data for the widget
    rules_list = [
        {
            "name": "Rule #1",
            "property": "Enable",
            "initial_value": "False",
            "expression": "ch[0] > 1",
            "channels": [{"channel": "ca://MTEST:Float", "trigger": True}],
        }
    ]
    # Add the rules to the widget
    widget.rules = json.dumps(rules_list)

    # Create a new Editor Window
    re = RulesEditor(widget)
    qtbot.addWidget(re)
    re.show()

    qtbot.waitExposed(re, timeout=5000)

    re.lst_rules.setCurrentRow(0)

    # make sure notes are saved and loaded correctly
    # in the scope of the current rules-editor window
    testing_text = "Testing adding note!"
    re.open_notes_window()
    re.notes_edit.setPlainText(testing_text)
    re.save_notes()
    re.open_notes_window()
    loaded_text_from_curr_editor_instance = re.notes_edit.toPlainText()
    assert loaded_text_from_curr_editor_instance == testing_text

    re.cancelChanges()



import logging
import pytest
import weakref

from ...widgets.rules import RulesDispatcher
from ...widgets.label import PyDMLabel


def test_rules_dispatcher(qapp, caplog):
    """
    Test the dispatcher to ensure that it is a singleton.

    Parameters
    ----------
    qapp : QApplication
        Reference to the QApplication
    caplog : fixture
        To capture the log messages
    """
    disp1 = RulesDispatcher()
    disp2 = RulesDispatcher()
    assert disp1 is disp2

    assert disp1.rules_engine.isRunning()

    payload = {"foo": "bar"}
    disp1.dispatch(payload)

    for record in caplog.records:
        assert record.levelno == logging.ERROR
    assert "Error at RulesDispatcher" in caplog.text


def test_unregister(qtbot):
    """
    Test the dispatcher for registering and unregistering of widgets.

    Parameters
    ----------
    qtbot : fixture
        Parent of all the widgets
    """
    widget = PyDMLabel()
    qtbot.addWidget(widget)

    rules = [
        {
            "name": "Rule #1",
            "property": "Visible",
            "expression": "ch[0] < 1",
            "channels": [{"channel": "ca://TESTRULES:Float", "trigger": True}],
        }
    ]

    dispatcher = RulesDispatcher()
    dispatcher.register(widget, rules)
    assert weakref.ref(widget) in dispatcher.rules_engine.widget_map

    dispatcher.unregister(widget)
    assert weakref.ref(widget) not in dispatcher.rules_engine.widget_map


def test_rules_not_connected(qtbot, caplog):
    """
    Test the rules mechanism.

    Parameters
    ----------
    qtbot : fixture
        Parent of all the widgets
    caplog : fixture
        To capture the log messages
    """
    widget = PyDMLabel()
    qtbot.addWidget(widget)
    widget.show()
    assert widget.isVisible()

    rules = [
        {
            "name": "Rule #1",
            "property": "Visible",
            "expression": "ch[0] < 1",
            "channels": [{"channel": "ca://TESTRULES:Float", "trigger": True}],
        }
    ]

    dispatcher = RulesDispatcher()
    dispatcher.register(widget, rules)

    re = dispatcher.rules_engine
    assert weakref.ref(widget) in re.widget_map
    assert len(re.widget_map[weakref.ref(widget)]) == 1
    assert re.widget_map[weakref.ref(widget)][0]["rule"] == rules[0]

    with caplog.at_level(logging.DEBUG):
        re.callback_value(weakref.ref(widget), 0, 0, trigger=True, value=1)
        for record in caplog.records:
            assert record.levelno == logging.DEBUG
        assert "Not all channels are connected" in caplog.text


def test_rules_ok(qtbot, caplog):
    """
    Test the rules mechanism.

    Parameters
    ----------
    qtbot : fixture
        Parent of all the widgets
    caplog : fixture
        To capture the log messages
    """

    widget = PyDMLabel()
    qtbot.addWidget(widget)
    widget.show()
    assert widget.isVisible()

    rules = [
        {
            "name": "Rule #1",
            "property": "Visible",
            "expression": "ch[0] < 1",
            "channels": [{"channel": "ca://TESTRULES:Float", "trigger": True}],
        }
    ]

    dispatcher = RulesDispatcher()
    dispatcher.register(widget, rules)

    re = dispatcher.rules_engine
    assert weakref.ref(widget) in re.widget_map
    assert len(re.widget_map[weakref.ref(widget)]) == 1
    assert re.widget_map[weakref.ref(widget)][0]["rule"] == rules[0]

    blocker = qtbot.waitSignal(re.rule_signal, timeout=1000)

    re.callback_conn(weakref.ref(widget), 0, 0, value=True)
    re.callback_value(weakref.ref(widget), 0, 0, trigger=True, value=5)
    assert re.widget_map[weakref.ref(widget)][0]["calculate"] is True

    blocker.wait()
    assert re.widget_map[weakref.ref(widget)][0]["calculate"] is False
    assert not widget.isVisible()


@pytest.mark.parametrize("use_enum, visible", [(None, True), (True, True), (False, False)])
def test_rules_enums(use_enum, visible, qtbot, caplog):
    """
    Test the rules mechanism with enums.

    Parameters
    ----------
    qtbot : fixture
        Parent of all the widgets
    caplog : fixture
        To capture the log messages
    """
    widget = PyDMLabel()
    qtbot.addWidget(widget)
    widget.show()
    assert widget.isVisible()

    rules = [
        {
            "name": "Rule #1",
            "property": "Visible",
            "expression": 'ch[0] == "RUN"',
            "channels": [{"channel": "ca://TESTRULES:Float", "trigger": True}],
        }
    ]
    if use_enum is not None:
        rules[0]["channels"][0]["use_enum"] = use_enum

    dispatcher = RulesDispatcher()
    dispatcher.register(widget, rules)

    re = dispatcher.rules_engine
    assert weakref.ref(widget) in re.widget_map
    assert len(re.widget_map[weakref.ref(widget)]) == 1
    assert re.widget_map[weakref.ref(widget)][0]["rule"] == rules[0]

    # First we test that we receive a value but we don't have enums yet
    blocker = qtbot.waitSignal(re.rule_signal, timeout=1000)

    re.callback_conn(weakref.ref(widget), 0, 0, value=True)
    re.callback_value(weakref.ref(widget), 0, 0, trigger=True, value=1)
    assert re.widget_map[weakref.ref(widget)][0]["calculate"] is True

    blocker.wait()
    assert re.widget_map[weakref.ref(widget)][0]["calculate"] is False
    assert not widget.isVisible()

    blocker = qtbot.waitSignal(re.rule_signal, timeout=1000)

    # Now receive enums and check that it was evaluated again and proper
    # value was sent making the widget visible on condition of use_enum
    re.callback_conn(weakref.ref(widget), 0, 0, value=True)
    re.callback_enum(weakref.ref(widget), 0, 0, enums=["STOP", "RUN"])
    assert re.widget_map[weakref.ref(widget)][0]["calculate"] is True

    blocker.wait()
    assert re.widget_map[weakref.ref(widget)][0]["calculate"] is False
    assert widget.isVisible() == visible


def test_rules_invalid_expr(qtbot, caplog):
    """
    Test the rules mechanism.

    Parameters
    ----------
    qtbot : fixture
        Parent of all the widgets
    caplog : fixture
        To capture the log messages
    """
    widget = PyDMLabel()
    qtbot.addWidget(widget)
    widget.show()
    assert widget.isVisible()

    rules = [
        {
            "name": "Rule #1",
            "property": "Visible",
            "expression": "ch[0] < 1",
            "channels": [{"channel": "ca://TESTRULES:Float", "trigger": True}],
        }
    ]

    dispatcher = RulesDispatcher()
    dispatcher.register(widget, rules)

    re = dispatcher.rules_engine
    assert weakref.ref(widget) in re.widget_map
    assert len(re.widget_map[weakref.ref(widget)]) == 1
    assert re.widget_map[weakref.ref(widget)][0]["rule"] == rules[0]

    caplog.clear()

    rules[0]["expression"] = "foo"
    dispatcher.register(widget, rules)
    assert len(re.widget_map[weakref.ref(widget)]) == 1
    re.callback_conn(weakref.ref(widget), 0, 0, value=True)
    re.callback_value(weakref.ref(widget), 0, 0, trigger=True, value="a")

    # Wait for rule to execute but keep app responsive
    qtbot.wait(1000)

    for record in caplog.records:
        assert record.levelno == logging.ERROR
    assert "Error while evaluating Rule" in caplog.text

    dispatcher.unregister(widget)
    assert weakref.ref(widget) not in re.widget_map


def test_rules_initial_value(qtbot, caplog):
    """
    Test the rules initial value mechanism.

    Parameters
    ----------
    qtbot : fixture
        Parent of all the widgets
    caplog : fixture
        To capture the log messages
    """
    widget = PyDMLabel()
    widget.setText("Defaut Label")
    qtbot.addWidget(widget)
    widget.show()

    rules = [
        {
            "name": "Rule #1",
            "property": "Text",
            "expression": "str(ch[0])",
            "initial_value": "Initial Value Test",
            "channels": [{"channel": "ca://TESTRULES:Float", "trigger": True}],
        }
    ]

    dispatcher = RulesDispatcher()
    dispatcher.register(widget, rules)

    re = dispatcher.rules_engine
    assert weakref.ref(widget) in re.widget_map
    assert len(re.widget_map[weakref.ref(widget)]) == 1
    assert re.widget_map[weakref.ref(widget)][0]["rule"] == rules[0]
    assert widget.text() == "Initial Value Test"
    blocker = qtbot.waitSignal(re.rule_signal, timeout=1000)

    re.callback_conn(weakref.ref(widget), 0, 0, value=True)
    re.callback_value(weakref.ref(widget), 0, 0, trigger=True, value=5)
    assert re.widget_map[weakref.ref(widget)][0]["calculate"] is True

    blocker.wait()
    assert re.widget_map[weakref.ref(widget)][0]["calculate"] is False
    assert widget.text() == str(5)



import pytest
from scipy import constants
from ...utilities import units


@pytest.mark.parametrize("typ, expected", [("cm", "length"), ("non_existent", None)])
def test_find_unittype(typ, expected):
    tp = units.find_unittype(typ)
    assert tp == expected


@pytest.mark.parametrize("unit, expected", [("cm", constants.centi), ("non_existent", None)])
def test_find_unit(unit, expected):
    r = units.find_unit(unit)
    assert r == expected


@pytest.mark.parametrize(
    "unit, desired, expected", [("m", "cm", 1 / constants.centi), ("m", "rad", None), ("non_existent", "rad", None)]
)
def test_convert(unit, desired, expected):
    r = units.convert(unit, desired)
    assert r == expected


@pytest.mark.parametrize("unit, expected", [("V", ["MV", "kV", "V", "mV", "uV"]), ("foo", None)])
def test_find_unit_options(unit, expected):
    opts = units.find_unit_options(unit)
    assert opts == expected



import os

__all__ = ["DEFAULT_PROTOCOL", "DESIGNER_ONLINE", "STYLESHEET", "STYLESHEET_INCLUDE_DEFAULT", "CONFIRM_QUIT"]


DEFAULT_PROTOCOL = os.getenv("PYDM_DEFAULT_PROTOCOL")
if DEFAULT_PROTOCOL is not None:
    # Get rid of the "://" part if it exists
    DEFAULT_PROTOCOL = DEFAULT_PROTOCOL.split("://")[0]

DESIGNER_ONLINE = os.getenv("PYDM_DESIGNER_ONLINE", None) is not None

STYLESHEET = os.getenv("PYDM_STYLESHEET", None)

STYLESHEET_INCLUDE_DEFAULT = os.getenv("PYDM_STYLESHEET_INCLUDE_DEFAULT", False)

CONFIRM_QUIT = os.getenv("PYDM_CONFIRM_QUIT", "n").lower() in ("y", "t", "1", "true")

# Environment variable pointing to a pydm display to return to when the home button is clicked
HOME_FILE = os.getenv("PYDM_HOME_FILE")

ENTRYPOINT_EXTERNAL_TOOL = "pydm.tool"
ENTRYPOINT_DATA_PLUGIN = "pydm.data_plugin"
ENTRYPOINT_WIDGET = "pydm.widget"

EXTERNAL_TOOL_SUFFIX = "_tool.py"
DATA_PLUGIN_SUFFIX = "_plugin.py"
